#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

# Implement a "failure filter" - that is, look at the output of a previous
# action and see if it failed with respect to downstream actions which need its
# outputs. This is to allow us to report success if rustc generated the artifact
# we needed (ie diagnostics) even if the compilation itself failed.

import argparse
import json
import os
import shutil
import sys
from typing import IO, List, NamedTuple, Optional, Tuple


class Args(NamedTuple):
    build_status: IO[str]
    required_file: Optional[List[Tuple[str, str, str]]]
    stderr: Optional[IO[str]]


def arg_parse() -> Args:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--build-status",
        type=argparse.FileType(),
        required=True,
    )
    parser.add_argument(
        "--required-file",
        action="append",
        nargs=3,
        metavar=("SHORT", "INPUT", "OUTPUT"),
    )
    parser.add_argument(
        "--stderr",
        type=argparse.FileType(),
    )

    return Args(**vars(parser.parse_args()))


def main() -> int:
    args = arg_parse()

    if args.stderr:
        stderr = args.stderr.read()
        sys.stderr.write(stderr)

    build_status = json.load(args.build_status)

    # Copy all required files to output, and fail with the original exit status
    # if any are missing. (Ideally we could just do the copy by referring to the
    # same underlying CAS object, which would avoid having to move the actual
    # bytes around at all.)
    if args.required_file:
        for short, inp, out in args.required_file:
            if short in build_status["files"]:
                try:
                    # Try a hard link to avoid unnecessary copies
                    os.link(inp, out)
                except OSError:
                    # Fall back to real copy if that doesn't work
                    shutil.copy(inp, out)
            else:
                print(
                    f"Missing required input file {short} ({inp})",
                    file=sys.stderr,
                )
                return build_status["status"]

    # If all the required files were present, then success regardless of
    # original status.
    return 0


sys.exit(main())



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Run `go cover` on non-`_test.go` input sources.
"""

# pyre-unsafe

import argparse
import hashlib
import subprocess
import sys
from pathlib import Path


def _var(pkg_name, src):
    return "Var_" + hashlib.md5(f"{pkg_name}::{src}".encode("utf-8")).hexdigest()


def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--cover", type=Path, required=True)
    parser.add_argument("--pkg-name", type=str, required=True)
    parser.add_argument("--coverage-mode", type=str, required=True)
    parser.add_argument("--covered-srcs-dir", type=Path, required=True)
    parser.add_argument("--out-srcs-argsfile", type=Path, required=True)
    parser.add_argument("--coverage-var-argsfile", type=Path, required=True)
    parser.add_argument("srcs", nargs="*", type=Path)
    args = parser.parse_args(argv[1:])

    out_srcs = []
    coverage_vars = {}

    args.covered_srcs_dir.mkdir(parents=True)

    for src in args.srcs:
        if src.name.endswith("_test.go"):
            out_srcs.append(src)
        else:
            var = _var(args.pkg_name, src)
            covered_src = args.covered_srcs_dir / src
            covered_src.parent.mkdir(parents=True, exist_ok=True)
            subprocess.check_call(
                [
                    args.cover,
                    "-mode",
                    args.coverage_mode,
                    "-var",
                    var,
                    "-o",
                    covered_src,
                    src,
                ]
            )
            # we need just the source name for the --cover-pkgs argument
            coverage_vars[var] = src.name
            out_srcs.append(covered_src)

    with open(args.out_srcs_argsfile, mode="w") as f:
        for src in out_srcs:
            print(src, file=f)

    with open(args.coverage_var_argsfile, mode="w") as f:
        if coverage_vars:
            print("--cover-pkgs", file=f)
            print(
                "{}:{}".format(
                    args.pkg_name,
                    ",".join([f"{var}={name}" for var, name in coverage_vars.items()]),
                ),
                file=f,
            )


sys.exit(main(sys.argv))



#!/usr/local/bin/python2.6 -tt
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
This file contains the main module code for buck python test programs.

By default, this is the main module for all python_test() rules.  However,
rules can also specify their own custom main_module.  If you write your own
main module, you can import this module as tools.test.stubs.fbpyunit, to access
any of its code to help implement your main module.
"""

# pyre-unsafe

from __future__ import print_function

import contextlib
import ctypes
import fnmatch
import json
import logging
import optparse
import os
import platform
import re
import sys
import time
import traceback
import unittest
import warnings


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import imp

try:
    from StringIO import StringIO  # type: ignore
except ImportError:
    from io import StringIO
try:
    import coverage  # type: ignore
except ImportError:
    coverage = None
try:
    from importlib.machinery import SourceFileLoader
except ImportError:
    SourceFileLoader = None


EXIT_CODE_SUCCESS = 0
EXIT_CODE_TEST_FAILURE = 70


class TestStatus(object):

    ABORTED = "FAILURE"
    PASSED = "SUCCESS"
    FAILED = "FAILURE"
    EXPECTED_FAILURE = "SUCCESS"
    UNEXPECTED_SUCCESS = "FAILURE"
    SKIPPED = "ASSUMPTION_VIOLATION"
    EXCLUDED = "EXCLUDED"


class PathMatcher(object):
    def __init__(self, include_patterns, omit_patterns):
        self.include_patterns = include_patterns
        self.omit_patterns = omit_patterns

    def omit(self, path):
        """
        Omit iff matches any of the omit_patterns or the include patterns are
        not empty and none is matched
        """
        path = os.path.realpath(path)
        return any(fnmatch.fnmatch(path, p) for p in self.omit_patterns) or (
            self.include_patterns
            and not any(fnmatch.fnmatch(path, p) for p in self.include_patterns)
        )

    def include(self, path):
        return not self.omit(path)


class DebugWipeFinder(object):
    """
    PEP 302 finder that uses a DebugWipeLoader for all files which do not need
    coverage
    """

    def __init__(self, matcher):
        self.matcher = matcher

    def find_module(self, fullname, path=None):
        _, _, basename = fullname.rpartition(".")
        try:
            fd, pypath, (_, _, kind) = imp.find_module(basename, path)
        except Exception:
            # Finding without hooks using the imp module failed. One reason
            # could be that there is a zip file on sys.path. The imp module
            # does not support loading from there. Leave finding this module to
            # the others finders in sys.meta_path.
            return None

        if hasattr(fd, "close"):
            fd.close()
        if kind != imp.PY_SOURCE:
            return None
        if self.matcher.include(pypath):
            return None

        """
        This is defined to match CPython's PyVarObject struct
        """

        class PyVarObject(ctypes.Structure):
            _fields_ = [
                ("ob_refcnt", ctypes.c_long),
                ("ob_type", ctypes.c_void_p),
                ("ob_size", ctypes.c_ulong),
            ]

        class DebugWipeLoader(SourceFileLoader):
            """
            PEP302 loader that zeros out debug information before execution
            """

            def get_code(self, fullname):
                code = super(DebugWipeLoader, self).get_code(fullname)
                if code:
                    # Ideally we'd do
                    # code.co_lnotab = b''
                    # But code objects are READONLY. Not to worry though; we'll
                    # directly modify CPython's object
                    code_impl = PyVarObject.from_address(id(code.co_lnotab))
                    code_impl.ob_size = 0
                return code

        return DebugWipeLoader(fullname, pypath)


def optimize_for_coverage(cov, include_patterns, omit_patterns):
    """
    We get better performance if we zero out debug information for files which
    we're not interested in. Only available in CPython 3.3+
    """
    matcher = PathMatcher(include_patterns, omit_patterns)
    if SourceFileLoader and platform.python_implementation() == "CPython":
        sys.meta_path.insert(0, DebugWipeFinder(matcher))


class TeeStream(object):
    def __init__(self, *streams):
        self._streams = streams

    def write(self, data):
        for stream in self._streams:
            stream.write(data)

    def flush(self):
        for stream in self._streams:
            stream.flush()

    def isatty(self):
        return False


class CallbackStream(object):
    def __init__(self, callback, bytes_callback=None, orig=None):
        self._callback = callback
        self._fileno = orig.fileno() if orig else None

        # Python 3 APIs:
        # - `encoding` is a string holding the encoding name
        # - `errors` is a string holding the error-handling mode for encoding
        # - `buffer` should look like an io.BufferedIOBase object

        self.errors = orig.errors if orig else None
        if bytes_callback:
            # those members are only on the io.TextIOWrapper
            self.encoding = orig.encoding if orig else "UTF-8"
            self.buffer = CallbackStream(bytes_callback, orig=orig)

    def write(self, data):
        self._callback(data)

    def flush(self):
        pass

    def isatty(self):
        return False

    def fileno(self):
        return self._fileno


# pyre-fixme[11]: Annotation `unittest._TextTestResult` is not defined as a type.
class BuckTestResult(unittest._TextTestResult):
    """
    Our own TestResult class that outputs data in a format that can be easily
    parsed by buck's test runner.
    """

    def __init__(
        self, stream, descriptions, verbosity, show_output, main_program, suite
    ):
        super(BuckTestResult, self).__init__(stream, descriptions, verbosity)
        self._main_program = main_program
        self._suite = suite
        self._results = []
        self._current_test = None
        self._saved_stdout = sys.stdout
        self._saved_stderr = sys.stderr
        self._show_output = show_output

    def getResults(self):
        return self._results

    def startTest(self, test):
        super(BuckTestResult, self).startTest(test)

        # Pass in the real stdout and stderr filenos.  We can't really do much
        # here to intercept callers who directly operate on these fileno
        # objects.
        sys.stdout = CallbackStream(
            self.addStdout, self.addStdoutBytes, orig=sys.stdout
        )
        sys.stderr = CallbackStream(
            self.addStderr, self.addStderrBytes, orig=sys.stderr
        )
        self._current_test = test
        self._test_start_time = time.time()
        self._current_status = TestStatus.ABORTED
        self._messages = []
        self._stacktrace = None
        self._stdout = ""
        self._stderr = ""

    def _find_next_test(self, suite):
        """
        Find the next test that has not been run.
        """

        for test in suite:

            # We identify test suites by test that are iterable (as is done in
            # the builtin python test harness).  If we see one, recurse on it.
            if hasattr(test, "__iter__"):
                test = self._find_next_test(test)

            # The builtin python test harness sets test references to `None`
            # after they have run, so we know we've found the next test up
            # if it's not `None`.
            if test is not None:
                return test

    def stopTest(self, test):
        sys.stdout = self._saved_stdout
        sys.stderr = self._saved_stderr

        super(BuckTestResult, self).stopTest(test)

        # If a failure occured during module/class setup, then this "test" may
        # actually be a `_ErrorHolder`, which doesn't contain explicit info
        # about the upcoming test.  Since we really only care about the test
        # name field (i.e. `_testMethodName`), we use that to detect an actual
        # test cases, and fall back to looking the test up from the suite
        # otherwise.
        if not hasattr(test, "_testMethodName"):
            test = self._find_next_test(self._suite)

        self._results.append(
            {
                "testCaseName": "{0}.{1}".format(
                    test.__class__.__module__, test.__class__.__name__
                ),
                "testCase": test._testMethodName,
                "type": self._current_status,
                "time": int((time.time() - self._test_start_time) * 1000),
                "message": os.linesep.join(self._messages),
                "stacktrace": self._stacktrace,
                "stdOut": self._stdout,
                "stdErr": self._stderr,
            }
        )

        self._current_test = None

    def stopTestRun(self):
        cov = self._main_program.get_coverage()
        if cov is not None:
            self._results.append({"coverage": cov})

    @contextlib.contextmanager
    def _withTest(self, test):
        self.startTest(test)
        yield
        self.stopTest(test)

    def _setStatus(self, test, status, message=None, stacktrace=None):
        assert test == self._current_test
        self._current_status = status
        self._stacktrace = stacktrace
        if message is not None:
            if message.endswith(os.linesep):
                message = message[:-1]
            self._messages.append(message)

    def setStatus(self, test, status, message=None, stacktrace=None):
        # addError() may be called outside of a test if one of the shared
        # fixtures (setUpClass/tearDownClass/setUpModule/tearDownModule)
        # throws an error.
        #
        # In this case, create a fake test result to record the error.
        if self._current_test is None:
            with self._withTest(test):
                self._setStatus(test, status, message, stacktrace)
        else:
            self._setStatus(test, status, message, stacktrace)

    def setException(self, test, status, excinfo):
        exctype, value, tb = excinfo
        self.setStatus(
            test,
            status,
            "{0}: {1}".format(exctype.__name__, value),
            "".join(traceback.format_tb(tb)),
        )

    def addSuccess(self, test):
        super(BuckTestResult, self).addSuccess(test)
        self.setStatus(test, TestStatus.PASSED)

    def addError(self, test, err):
        super(BuckTestResult, self).addError(test, err)
        self.setException(test, TestStatus.ABORTED, err)

    def addFailure(self, test, err):
        super(BuckTestResult, self).addFailure(test, err)
        self.setException(test, TestStatus.FAILED, err)

    def addSkip(self, test, reason):
        super(BuckTestResult, self).addSkip(test, reason)
        self.setStatus(test, TestStatus.SKIPPED, "Skipped: %s" % (reason,))

    def addExpectedFailure(self, test, err):
        super(BuckTestResult, self).addExpectedFailure(test, err)
        self.setException(test, TestStatus.EXPECTED_FAILURE, err)

    def addUnexpectedSuccess(self, test):
        super(BuckTestResult, self).addUnexpectedSuccess(test)
        self.setStatus(test, TestStatus.UNEXPECTED_SUCCESS, "Unexpected success")

    def addStdout(self, val):
        self._stdout += val
        if self._show_output:
            self._saved_stdout.write(val)
            self._saved_stdout.flush()

    def addStdoutBytes(self, val):
        string = val.decode("utf-8", errors="backslashreplace")
        self.addStdout(string)

    def addStderr(self, val):
        self._stderr += val
        if self._show_output:
            self._saved_stderr.write(val)
            self._saved_stderr.flush()

    def addStderrBytes(self, val):
        string = val.decode("utf-8", errors="backslashreplace")
        self.addStderr(string)


class BuckTestRunner(unittest.TextTestRunner):
    def __init__(self, main_program, suite, show_output=True, **kwargs):
        super(BuckTestRunner, self).__init__(**kwargs)
        self.show_output = show_output
        self._main_program = main_program
        self._suite = suite

    def _makeResult(self):
        return BuckTestResult(
            self.stream,
            self.descriptions,
            self.verbosity,
            self.show_output,
            self._main_program,
            self._suite,
        )


def _format_test_name(test_class, attrname):
    """
    Format the name of the test buck-style.
    """
    return "{0}.{1}#{2}".format(test_class.__module__, test_class.__name__, attrname)


class StderrLogHandler(logging.StreamHandler):
    """
    This class is very similar to logging.StreamHandler, except that it
    always uses the current sys.stderr object.

    StreamHandler caches the current sys.stderr object when it is constructed.
    This makes it behave poorly in unit tests, which may replace sys.stderr
    with a StringIO buffer during tests.  The StreamHandler will continue using
    the old sys.stderr object instead of the desired StringIO buffer.
    """

    def __init__(self):
        logging.Handler.__init__(self)

    @property
    def stream(self):
        return sys.stderr


class RegexTestLoader(unittest.TestLoader):
    def __init__(self, regex=None):
        self.regex = regex
        super(RegexTestLoader, self).__init__()

    def getTestCaseNames(self, testCaseClass):
        """
        Return a sorted sequence of method names found within testCaseClass
        """

        testFnNames = super(RegexTestLoader, self).getTestCaseNames(testCaseClass)
        if self.regex is None:
            return testFnNames
        robj = re.compile(self.regex)
        matched = []
        for attrname in testFnNames:
            fullname = _format_test_name(testCaseClass, attrname)
            if robj.search(fullname):
                matched.append(attrname)
        return matched


class Loader(object):
    def __init__(self, modules, regex=None):
        self.modules = modules
        self.regex = regex

    def load_all(self):
        loader = RegexTestLoader(self.regex)
        test_suite = unittest.TestSuite()
        for module_name in self.modules:
            __import__(module_name, level=0)
            module = sys.modules[module_name]
            module_suite = loader.loadTestsFromModule(module)
            test_suite.addTest(module_suite)
        return test_suite

    def load_args(self, args):
        loader = RegexTestLoader(self.regex)

        suites = []
        for arg in args:
            suite = loader.loadTestsFromName(arg)
            # loadTestsFromName() can only process names that refer to
            # individual test functions or modules.  It can't process package
            # names.  If there were no module/function matches, check to see if
            # this looks like a package name.
            if suite.countTestCases() != 0:
                suites.append(suite)
                continue

            # Load all modules whose name is <arg>.<something>
            prefix = arg + "."
            for module in self.modules:
                if module.startswith(prefix):
                    suite = loader.loadTestsFromName(module)
                    suites.append(suite)

        return loader.suiteClass(suites)


class MainProgram(object):
    """
    This class implements the main program.  It can be subclassed by
    users who wish to customize some parts of the main program.
    (Adding additional command line options, customizing test loading, etc.)
    """

    DEFAULT_VERBOSITY = 2

    def __init__(self, argv):
        self.init_option_parser()
        self.parse_options(argv)
        self.setup_logging()

    def init_option_parser(self):
        usage = "%prog [options] [TEST] ..."
        op = optparse.OptionParser(usage=usage, add_help_option=False)
        self.option_parser = op

        op.add_option(
            "--hide-output",
            dest="show_output",
            action="store_false",
            default=True,
            help="Suppress data that tests print to stdout/stderr, and only "
            "show it if the test fails.",
        )
        op.add_option(
            "-o",
            "--output",
            help="Write results to a file in a JSON format to be read by Buck",
        )
        op.add_option(
            "-f",
            "--failfast",
            action="store_true",
            default=False,
            help="Stop after the first failure",
        )
        op.add_option(
            "-l",
            "--list-tests",
            action="store_true",
            dest="list",
            default=False,
            help="List tests and exit",
        )
        op.add_option(
            "-L",
            "--list-format",
            dest="list_format",
            choices=["buck", "python"],
            default="python",
            help="List tests format",
        )
        op.add_option(
            "-r",
            "--regex",
            default=None,
            help="Regex to apply to tests, to only run those tests",
        )
        op.add_option(
            "--collect-coverage",
            action="store_true",
            default=False,
            help="Collect test coverage information",
        )
        op.add_option(
            "--coverage-include",
            default="*",
            help='File globs to include in converage (split by ",")',
        )
        op.add_option(
            "--coverage-omit",
            default="",
            help='File globs to omit from converage (split by ",")',
        )
        op.add_option(
            "--logger",
            action="append",
            metavar="<category>=<level>",
            default=[],
            help="Configure log levels for specific logger categories",
        )
        op.add_option(
            "-q",
            "--quiet",
            action="count",
            default=0,
            help="Decrease the verbosity (may be specified multiple times)",
        )
        op.add_option(
            "-v",
            "--verbosity",
            action="count",
            default=self.DEFAULT_VERBOSITY,
            help="Increase the verbosity (may be specified multiple times)",
        )
        op.add_option(
            "-?", "--help", action="help", help="Show this help message and exit"
        )

    def parse_options(self, argv):
        self.options, self.test_args = self.option_parser.parse_args(argv[1:])
        self.options.verbosity -= self.options.quiet

        if self.options.collect_coverage and coverage is None:
            self.option_parser.error("coverage module is not available")
        self.options.coverage_include = self.options.coverage_include.split(",")
        if self.options.coverage_omit == "":
            self.options.coverage_omit = []
        else:
            self.options.coverage_omit = self.options.coverage_omit.split(",")

    def setup_logging(self):
        # Configure the root logger to log at INFO level.
        # This is similar to logging.basicConfig(), but uses our
        # StderrLogHandler instead of a StreamHandler.
        fmt = logging.Formatter("%(pathname)s:%(lineno)s: %(message)s")
        log_handler = StderrLogHandler()
        log_handler.setFormatter(fmt)
        root_logger = logging.getLogger()
        root_logger.addHandler(log_handler)
        root_logger.setLevel(logging.INFO)

        level_names = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warn": logging.WARNING,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL,
            "fatal": logging.FATAL,
        }

        for value in self.options.logger:
            parts = value.rsplit("=", 1)
            if len(parts) != 2:
                self.option_parser.error(
                    "--logger argument must be of the "
                    "form <name>=<level>: %s" % value
                )
            name = parts[0]
            level_name = parts[1].lower()
            level = level_names.get(level_name)
            if level is None:
                self.option_parser.error(
                    "invalid log level %r for log category %s" % (parts[1], name)
                )
            logging.getLogger(name).setLevel(level)

    def create_loader(self):
        import __test_modules__

        return Loader(__test_modules__.TEST_MODULES, self.options.regex)

    def load_tests(self):
        loader = self.create_loader()
        if self.options.collect_coverage:
            self.start_coverage()
            include = self.options.coverage_include
            omit = self.options.coverage_omit
            if include and "*" not in include:
                optimize_for_coverage(self.cov, include, omit)

        if self.test_args:
            suite = loader.load_args(self.test_args)
        else:
            suite = loader.load_all()
        if self.options.collect_coverage:
            self.cov.start()
        return suite

    def get_tests(self, test_suite):
        tests = []

        for test in test_suite:
            if isinstance(test, unittest.TestSuite):
                tests.extend(self.get_tests(test))
            else:
                tests.append(test)

        return tests

    def run(self):
        test_suite = self.load_tests()

        if self.options.list:
            for test in self.get_tests(test_suite):
                if self.options.list_format == "python":
                    name = str(test)
                elif self.options.list_format == "buck":
                    method_name = getattr(test, "_testMethodName", "")
                    name = _format_test_name(test.__class__, method_name)
                else:
                    raise Exception(
                        "Bad test list format: %s" % (self.options.list_format,)
                    )

                print(name)
            return EXIT_CODE_SUCCESS
        else:
            result = self.run_tests(test_suite)
            if self.options.output is not None:
                with open(self.options.output, "w") as f:
                    json.dump(result.getResults(), f, indent=4, sort_keys=True)
            if not result.wasSuccessful():
                return EXIT_CODE_TEST_FAILURE
            return EXIT_CODE_SUCCESS

    def run_tests(self, test_suite):
        # Install a signal handler to catch Ctrl-C and display the results
        # (but only if running >2.6).
        if sys.version_info[0] > 2 or sys.version_info[1] > 6:
            unittest.installHandler()

        # Run the tests
        runner = BuckTestRunner(
            self,
            test_suite,
            verbosity=self.options.verbosity,
            show_output=self.options.show_output,
        )
        result = runner.run(test_suite)

        if self.options.collect_coverage and self.options.show_output:
            self.cov.stop()
            if self.cov.html_report:
                self.cov.html_report()
            else:
                self.cov.report(file=sys.stdout)

        return result

    def start_coverage(self):
        if not self.options.collect_coverage:
            return

        # Keep the original working dir in case tests use os.chdir
        self._original_working_dir = os.getcwd()

        self.cov = coverage.Coverage(
            include=self.options.coverage_include, omit=self.options.coverage_omit
        )
        self.cov.erase()
        self.cov.start()

    def get_coverage(self):
        if not self.options.collect_coverage:
            return None
        result = {}

        # Switch back to the original working directory.
        os.chdir(self._original_working_dir)

        self.cov.stop()

        try:
            f = StringIO()
            self.cov.report(file=f)
            lines = f.getvalue().split("\n")
        except coverage.misc.CoverageException:
            # Nothing was covered. That's fine by us
            return result

        # N.B.: the format of the coverage library's output differs
        # depending on whether one or more files are in the results
        for line in lines[2:]:
            if line.strip("-") == "":
                break
            r = line.split()[0]
            analysis = self.cov.analysis2(r)
            covString = self.convert_to_diff_cov_str(analysis)
            if covString:
                result[r] = covString

        return result

    def convert_to_diff_cov_str(self, analysis):
        # Info on the format of analysis:
        # http://nedbatchelder.com/code/coverage/api.html
        if not analysis:
            return None
        numLines = max(
            analysis[1][-1] if len(analysis[1]) else 0,
            analysis[2][-1] if len(analysis[2]) else 0,
            analysis[3][-1] if len(analysis[3]) else 0,
        )
        lines = ["N"] * numLines
        for l in analysis[1]:
            lines[l - 1] = "C"
        for l in analysis[2]:
            lines[l - 1] = "X"
        for l in analysis[3]:
            lines[l - 1] = "U"
        return "".join(lines)


def main(argv):
    return MainProgram(sys.argv).run()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Prepares for an object-only ThinLTO link by extracting a given archive and
producing a manifest of the objects contained within.
"""

import argparse
import enum
import json
import os
import subprocess
import sys
from typing import List, Tuple


class ArchiveKind(enum.IntEnum):
    UNKNOWN = 0
    ARCHIVE = 1
    THIN_ARCHIVE = 2


def _gen_path(parent_path: str, filename: str) -> str:
    # concat file path and check the file exist before return
    obj_path = os.path.join(parent_path, filename)
    assert os.path.exists(obj_path)
    return obj_path


def _gen_filename(filename: str, num_of_instance: int) -> str:
    # generate the filename based on the instance,
    # for 1st instance, it's file.o
    # for 2nd instance, it's file_1.o
    if num_of_instance > 1:
        basename, extension = os.path.splitext(filename)
        return f"{basename}_{num_of_instance-1}{extension}"
    else:
        return filename


def identify_file(path: str) -> Tuple[ArchiveKind, str]:
    path = os.path.realpath(path)
    output = subprocess.check_output(["file", "-b", path]).decode()
    if "ar archive" in output:
        return (ArchiveKind.ARCHIVE, output)
    elif "thin archive" in output:
        return (ArchiveKind.THIN_ARCHIVE, output)
    else:
        with open(path, "rb") as infile:
            head = infile.read(7)

            if head == "!<thin>".encode():
                return (ArchiveKind.THIN_ARCHIVE, output)

    return (ArchiveKind.UNKNOWN, output)


def main(argv: List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--manifest-out")
    parser.add_argument("--objects-out")
    parser.add_argument("--ar")
    parser.add_argument("--name")
    parser.add_argument("--archive")
    args = parser.parse_args(argv[1:])

    objects_path = args.objects_out
    os.makedirs(objects_path, exist_ok=True)

    known_objects = []
    file_type, debug_output = identify_file(args.archive)
    if file_type == ArchiveKind.ARCHIVE:
        # Unfortunately, we use llvm-ar and, while binutils ar has had --output for
        # a long time, llvm-ar does not support --output and the change in llvm-ar
        # looks like it has stalled for years (https://reviews.llvm.org/D69418)
        # So, we need to invoke ar in the directory that we want it to extract into, and so
        # need to adjust some paths.
        ar_path = os.path.relpath(args.ar, start=objects_path)
        archive_path = os.path.relpath(args.archive, start=objects_path)
        output = subprocess.check_output(
            [ar_path, "t", archive_path], cwd=objects_path
        ).decode()
        member_list = [member for member in output.split("\n") if member]

        if len(set(member_list)) != len(member_list):
            # duplication detected
            counter = {}

            for member in member_list:
                current = counter.get(member, 0) + 1
                counter[member] = current
                if current == 1:
                    # just extract the file
                    output = subprocess.check_output(
                        [ar_path, "xN", str(current), archive_path, member],
                        cwd=objects_path,
                    ).decode()
                    assert not output
                    known_objects.append(_gen_path(objects_path, member))
                else:
                    # llvm doesn't allow --output so we need this clumsiness
                    tmp_filename = "tmp"
                    current_file = _gen_filename(member, current)
                    # rename current 'member' file to tmp
                    output = subprocess.check_output(
                        ["mv", member, tmp_filename], cwd=objects_path
                    ).decode()
                    assert not output
                    # extract the file from archive
                    output = subprocess.check_output(
                        [ar_path, "xN", str(current), archive_path, member],
                        cwd=objects_path,
                    ).decode()
                    assert not output
                    # rename the newly extracted file
                    output = subprocess.check_output(
                        ["mv", member, current_file], cwd=objects_path
                    ).decode()
                    assert not output
                    # rename the tmp file back to 'member'
                    output = subprocess.check_output(
                        ["mv", tmp_filename, member], cwd=objects_path
                    ).decode()
                    assert not output
                    known_objects.append(_gen_path(objects_path, current_file))
        else:
            # no duplicated filename
            output = subprocess.check_output(
                [ar_path, "xv", archive_path], cwd=objects_path
            ).decode()
            for line in output.splitlines():
                assert line.startswith("x - ")
                obj = line[4:]
                known_objects.append(_gen_path(objects_path, obj))

    elif file_type == ArchiveKind.THIN_ARCHIVE:
        output = subprocess.check_output([args.ar, "t", args.archive]).decode()
        for line in output.splitlines():
            assert os.path.exists(line)
            known_objects.append(line)
    elif file_type == ArchiveKind.UNKNOWN:
        raise AssertionError(
            f"unknown archive kind for file {args.archive}: {debug_output}"
        )

    manifest = {
        "debug": debug_output,
        "objects": known_objects,
    }
    with open(os.path.join(args.manifest_out), "w") as outfile:
        json.dump(manifest, outfile, indent=2, sort_keys=True)

    return 0


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

# Wrapper for rustc (or similar, like rustdoc). This wrapper does a few pieces
# of post-processing on the json-formatted diagnostics:
# - (preprocessing) resolve env vars referring to paths to absolute paths
# - write the rendered form to a text diagnostic output, and also to stderr
# - annotate unused crate messages with buck target info for downstream tooling,
#   and also generated a rendered version
# - generate a build status json when using failure filtering
#
# This is closely coupled to `_rustc_invoke` in `build.bzl`

import argparse
import asyncio
import json
import os
import platform
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Dict, IO, List, NamedTuple, Optional, Tuple

DEBUG = False


def key_value_arg(s: str) -> Tuple[str, str]:
    key_value = s.split("=", maxsplit=1)
    if len(key_value) == 2:
        return (key_value[0], key_value[1])
    raise argparse.ArgumentTypeError(f"expected the form `key=value` for `{s}`")


class Args(NamedTuple):
    diag_json: Optional[IO[str]]
    diag_txt: Optional[IO[str]]
    env: Optional[List[Tuple[str, str]]]
    path_env: Optional[List[Tuple[str, str]]]
    remap_cwd_prefix: Optional[str]
    crate_map: Optional[List[Tuple[str, str]]]
    buck_target: Optional[str]
    failure_filter: Optional[IO[str]]
    required_output: Optional[List[Tuple[str, str]]]
    rustc: List[str]


def arg_parse() -> Args:
    # Command line is <action.py> [args] -- rustc command line
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument(
        "--diag-json",
        type=argparse.FileType("w"),
        help="Json-formatted diagnostic output "
        "(assumes compiler is invoked with --error-format=json)",
    )
    parser.add_argument(
        "--diag-txt",
        type=argparse.FileType("w"),
        help="Rendered text diagnostic output (also streamed to stderr)",
    )
    parser.add_argument(
        "--env",
        action="append",
        type=key_value_arg,
        metavar="NAME=VALUE",
        help="Set environment",
    )
    parser.add_argument(
        "--path-env",
        action="append",
        type=key_value_arg,
        metavar="NAME=PATH",
        help="Set path environment (to be made absolute)",
    )
    parser.add_argument(
        "--remap-cwd-prefix",
        help="Remap paths under the current working directory to this path prefix",
    )
    parser.add_argument(
        "--crate-map",
        action="append",
        type=key_value_arg,
        metavar="CRATE=TARGET",
        help="Crate name to target map for unused crate diagnostics",
    )
    parser.add_argument(
        "--buck-target",
        help="Buck target for crate, used for unused crate diagnostics",
    )
    parser.add_argument(
        "--failure-filter",
        type=argparse.FileType(mode="w"),
        help="Consider a failure as success so long as we got some usable diagnostics",
        metavar="build-status.json",
    )
    parser.add_argument(
        "--required-output",
        action="append",
        nargs=2,
        metavar=("SHORT", "PATH"),
        help="Required output path we expect rustc to generate "
        "(and filled with a placeholder on a filtered failure)",
    )
    parser.add_argument(
        "rustc",
        nargs=argparse.REMAINDER,
        type=str,
        help="Compiler command line",
    )

    return Args(**vars(parser.parse_args()))


async def handle_output(  # noqa: C901
    proc: asyncio.subprocess.Process,
    args: Args,
    crate_map: Dict[str, str],
) -> bool:
    got_error_diag = False

    proc_stderr = proc.stderr
    assert proc_stderr is not None

    while True:
        line = await proc_stderr.readline()

        if line is None or line == b"":
            break

        try:
            diag = json.loads(line)
        except json.JSONDecodeError:
            sys.stderr.buffer.write(line + b"\n")  # Passthrough
            continue

        if DEBUG:
            print(f"diag={repr(diag)}")

        if diag.get("level") == "error":
            got_error_diag = True

        # Add more information to unused crate warnings
        unused_names = diag.get("unused_extern_names", None)
        if unused_names:
            # Treat error-level unused dep warnings as errors
            if diag.get("lint_level") in ("deny", "forbid"):
                got_error_diag = True

            if args.buck_target:
                rendered_unused = []
                for name in unused_names:
                    if name in crate_map:
                        rendered_unused.append("{}: {}".format(crate_map[name], name))
                    else:
                        rendered_unused.append("{}".format(name))
                rendered_unused.sort()
                rendered_unused = "\n    ".join(rendered_unused)

                diag["buck_target"] = args.buck_target
                diag["rendered"] = (
                    f"Target `{args.buck_target}` has unused dependencies:\n"
                    f"    {rendered_unused}"
                )
            diag["unused_deps"] = {
                name: crate_map[name] for name in unused_names if name in crate_map
            }

        # Emit json
        if args.diag_json:
            args.diag_json.write(json.dumps(diag, separators=(",", ":")) + "\n")

        # Emit rendered text version
        if "rendered" in diag:
            rendered = diag["rendered"] + "\n"
            if args.diag_txt:
                args.diag_txt.write(rendered)
            sys.stderr.write(rendered)

    return got_error_diag


async def main() -> int:
    args = arg_parse()

    # Inherit a very limited initial environment, then add the new things
    env = {
        k: os.environ[k]
        for k in [
            "RUSTC_LOG",
            "RUST_BACKTRACE",
            "PATH",
            "PWD",
            "HOME",
            "TMPDIR",
            # Required on Windows
            "TEMP",
            # TODO(andirauter): Required by RE. Remove them when no longer required T119466023
            "EXECUTION_ID",
            "SESSION_ID",
            "CAS_DAEMON_PORT",
            # Required by Dotslash, which is how the Rust toolchain is shipped
            # on Mac.
            "USER",
            "DOTSLASH_CACHE",
        ]
        if k in os.environ
    }
    if args.env:
        env.update(dict(args.env))
    if args.path_env:
        env.update({k: str(Path(v).resolve()) for k, v in args.path_env})

    crate_map = dict(args.crate_map) if args.crate_map else {}

    if DEBUG:
        print(f"args {repr(args)} env {env} crate_map {crate_map}")

    # `tools/build_defs/rustc_plugin_platform.bzl` hardcodes building
    # proc-macros for x86-64 because we have no way of knowing what
    # architecture the rustc using them is going to eventually run on.
    # Since it's possible that we're on Apple Silicon, force rustc
    # to run under Rosetta so proc-macros are compatible.
    if sys.platform == "darwin" and platform.machine() == "arm64":
        rustc_cmd = ["arch", "-x86_64"] + args.rustc
    else:
        rustc_cmd = args.rustc

    if args.remap_cwd_prefix is not None:
        rustc_cmd.append(
            "--remap-path-prefix={}={}".format(os.getcwd(), args.remap_cwd_prefix)
        )

    # Kick off the action
    proc = await asyncio.create_subprocess_exec(
        *rustc_cmd,
        env=env,
        stdin=subprocess.DEVNULL,
        stdout=None,  # Inherit
        stderr=subprocess.PIPE,
        limit=1_000_000,
    )

    got_error_diag = await handle_output(proc, args, crate_map)

    res = await proc.wait()
    if DEBUG:
        print(
            f"res={repr(res)} "
            f"got_error_diag={got_error_diag} "
            f"args.failure_filter {args.failure_filter}"
        )

    # If rustc is reporting a silent error, make it loud
    if res == 0 and got_error_diag:
        res = 1

    # Check for death by signal - this is always considered a failure
    if res < 0:
        cmdline = " ".join(shlex.quote(arg) for arg in args.rustc)
        print(
            f"Command exited with signal {-res}: command line: {cmdline}",
            file=sys.stderr,
        )
    elif args.failure_filter:
        # If failure filtering is enabled, then getting an error diagnostic is also
        # considered a success. That is, if rustc exited with an error status, but
        # we saw a fatal error diagnostic, we can still report a zero exit status.
        # This still means we had an error if someone wanted one of the other output
        # artifacts as an input, but the failure filter action will handle that by
        # examining the build_status json output.

        required_output = args.required_output or []

        # We need to create a build status output, including the list of output
        # files which were *actuallyq* created. We use the short paths as the
        # logical filename rather than the actual full path, since that won't
        # mean much to a downstream action.
        build_status = {
            "status": res,
            "files": [short for short, path in required_output if Path(path).exists()],
        }
        json.dump(build_status, args.failure_filter)

        # OK to actually report success, but keep buck happy by making sure all
        # the required outputs are present
        if got_error_diag and res != 0:
            for _short, path in required_output:
                path = Path(path)
                if not path.exists():
                    path.touch()
            res = 0

    return res


sys.exit(asyncio.run(main()))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Quick and dirty wrapper to extract zip files; python 3.6.2+

extract.py my_zip_file.zip --output=output_directory
"""

import argparse
import shutil
from pathlib import Path


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Extract .zip files to a directory in a cross platform manner"
    )
    parser.add_argument(
        "--output", type=Path, required=True, help="The directory to write to"
    )
    parser.add_argument("src", type=Path, help="The archive to extract to --output")
    args = parser.parse_args()

    args.output.mkdir(parents=True, exist_ok=True)
    shutil.unpack_archive(args.src, args.output, "zip")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Creates a Python Source DB JSON file containing both a rule's immediate sources
and the sources of all transitive dependencies (e.g. for use with Pyre).

Sources and dependencies are passed in via source manifest files, which are
merged by this script:

$ ./make_source_db.py \
      --sources my_rule.manifest.json \
      --dependency dep1.manifest.json \
      --dependency dep2.manifest.json

The output format of the source DB is:

{
  "sources": {
    <source1-name>: <source1-path>,
    <source2-name>: <source2-path>,
    ...
  },
  "dependencies": {
    <dep-source1-name>: <dep-source1-path>,
    <dep-source2-name>: <dep-source2-path>,
    ...
  },
}
"""

# pyre-unsafe

import argparse
import json
import sys


def _load(path):
    with open(path) as f:
        return json.load(f)


def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("--sources")
    parser.add_argument("--dependency", action="append", default=[])
    args = parser.parse_args(argv[1:])

    db = {}

    # Add sources.
    sources = {}
    if args.sources is not None:
        for name, path, _ in _load(args.sources):
            sources[name] = path
    db["sources"] = sources

    # Add dependencies.
    dependencies = {}
    for dep in args.dependency:
        for name, path, origin in _load(dep):
            prev = dependencies.get(name)
            if prev is not None and prev[0] != path:
                raise Exception(
                    "Duplicate entries for {}: {} ({}) and {} ({})".format(
                        name, path, origin, *prev
                    ),
                )
            dependencies[name] = path, origin
    db["dependencies"] = {n: p for n, (p, _) in dependencies.items()}

    # Write db out.
    json.dump(db, args.output, indent=2)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Python wrapper around `clang` intended for use by the parallel opt phase of
a Distributed ThinLTO compilation. This script works around a LLVM bug where
LLVM will return a zero exit code in the case where ThinLTO fails with a
fatal error.

Instead of trusting the exit code of the compiler, this script checks the
output file and returns 1 if the file has zero size.
"""

import argparse
import os
import subprocess
import sys
from typing import List

EXIT_SUCCESS, EXIT_FAILURE = 0, 1

# Filter opt related flags
def _filter_flags(clang_flags: List[str]) -> List[str]:  # noqa: C901
    # List of llvm flags to be ignored.
    # They either don't have an valid mapping or unused during opt.
    IGNORE_OPT_FLAGS = [
        "-Wl,-plugin-opt,-function-sections",
        "-Wl,--lto-whole-program-visibility",
        "-Wl,--no-lto-whole-program-visibility",
    ]
    # Conservatively, we only translate llvms flags in our known list
    KNOWN_LLVM_SHARED_LIBRARY_FLAGS = ["-shared"]

    # Start with default flags for opt.
    # The default values may change across compiler versions.
    # Make sure they are always synced with the current values.
    opt_flags = [
        # TODO(T139459294):
        # -O2 is the default optimization flag for the link-time optimizer
        # this setting matches current llvm implementation:
        # https://github.com/llvm/llvm-project/blob/main/llvm/include/llvm/LTO/Config.h#L57
        "-O2",
        # TODO(T139459170): Remove after clang-15. NPM is the default.
        "-fexperimental-new-pass-manager",
        "-ffunction-sections",
        "-fdata-sections",
    ]

    # Clang driver passes through lld flags with "-Wl," prefix. There are 4 type of flags with unique
    # prefixes:
    # 1. "--lto-...": these are native lld flags.
    # 2. "-plugin-opt,..." or "-plugin-opt=...": these are the aliases of the native lld flags (1).
    # 3. "-mllvm,...": these are llvm flags.
    # 4. "-plugin-opt,-..." or "-plugin-opt=-...": these are the aliases of llvm flags (3). Note that they differ from (2) and always start with "-".
    #
    # For (1) and (2), we need to convert them case by case.
    # For (3) and (4), we should be able to pass them through into the optimizer directly by prefixing "-mllvm".
    # TODO(T139448744): Cover all the flags. Check available flags using "ld.lld --help | grep -A 1 '\-\-plugin-opt='"
    PLUGIN_OPT_PREFIXES = ["-Wl,-plugin-opt,", "-Wl,-plugin-opt="]

    def _find_plugin_opt_prefix(flag: str) -> str:
        matched_prefix = [
            prefix for prefix in PLUGIN_OPT_PREFIXES if flag.startswith(prefix)
        ]
        if matched_prefix:
            return matched_prefix[0]
        return ""

    plugin_opt_to_llvm_flag_map = {
        "sample-profile=": "-fprofile-sample-use=",
        "O": "-O",
    }

    def _plugin_opt_to_clang_flag(flag: str) -> str:
        for k, v in plugin_opt_to_llvm_flag_map.items():
            if flag.startswith(k):
                return flag.replace(k, v)
        return None

    for raw_flag in clang_flags:
        flag = raw_flag.replace('"', "")
        if flag in IGNORE_OPT_FLAGS:
            continue
        if _find_plugin_opt_prefix(flag):
            # Convert "-Wl,-plugin-opt,...".
            flag = flag.replace(_find_plugin_opt_prefix(flag), "", 1)
            if flag.startswith("-"):
                # If flag starts with "-", it is an llvm flag. Pass it through directly.
                opt_flags.extend(["-mllvm", flag])
            else:
                flag = _plugin_opt_to_clang_flag(flag)
                if flag is None:
                    # Bail on any unknown flag.
                    print(f"error: unrecognized flag {raw_flag}")
                    return None
                opt_flags.append(flag)
        elif flag.startswith("-Wl,-mllvm,"):
            # Convert "-Wl,-mllvm,...". It is an llvm flag. Pass it through directly.
            flag = flag.replace("-Wl,-mllvm,", "", 1)
            opt_flags.extend(["-mllvm", flag])
        elif flag in KNOWN_LLVM_SHARED_LIBRARY_FLAGS:
            # The target is a shared library, `-fPIC` is needed in opt phase to correctly generate PIC ELF.
            opt_flags.append("-fPIC")

    return opt_flags


# Clean up clang flags by obtaining the cc1 flags and filtering out those unwanted.
# clang_opt_flags is mutated after calling this function.
def _cleanup_flags(clang_opt_flags: List[str]) -> List[str]:
    for i, arg in enumerate(clang_opt_flags):
        if arg.startswith("--cc="):
            # Find the clang binary path.
            clang_opt_flags[i] = arg.replace("--cc=", "")
            break

    # Get the cc1 flag dump with '-###'
    try:
        output = (
            subprocess.check_output(
                clang_opt_flags + ["-###"], stderr=subprocess.STDOUT
            )
            .decode()
            .splitlines()
        )
    except subprocess.CalledProcessError as e:
        print(e.output.decode())
        return None

    # Flags that may conflict with the existing bitcode attributes.
    # The value indicates if the flag is followed with a value.
    flags_to_delete = {
        "-mframe-pointer=none": False,
        "-fmath-errno": False,
        "-fno-rounding-math": False,
        "-mconstructor-aliases": False,
        "-munwind-tables": False,
        "-target-cpu": True,
        "-tune-cpu": True,
    }

    clean_output = []
    skip_next = False
    for f in output[-1].split()[1:]:
        if skip_next:
            skip_next = False
        else:
            f = f.strip('"')
            if f in flags_to_delete:
                skip_next = flags_to_delete[f]
            else:
                clean_output.append(f)
    return clean_output


def main(argv: List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--out", help="The output native object file.")
    parser.add_argument("--input", help="The input bitcode object file.")
    parser.add_argument("--index", help="The thinlto index file.")
    parser.add_argument("--split-dwarf", required=False, help="Split dwarf option.")
    parser.add_argument(
        "--args", help="The argsfile containing unfiltered and unprocessed flags."
    )
    parser.add_argument("--debug", action="store_true", help="Dump clang -cc1 flags.")
    parser.add_argument("opt_args", nargs=argparse.REMAINDER)
    args = parser.parse_args(argv[1:])

    with open(args.args, "r") as argsfile:
        clang_opt_flags = _filter_flags(argsfile.read().splitlines())
    if clang_opt_flags is None:
        return EXIT_FAILURE

    clang_opt_flags.extend(
        [
            "-o",
            args.out,
            "-x",
            "ir",
            "-c",
            args.input,
            f"-fthinlto-index={args.index}",
        ]
    )
    if args.split_dwarf:
        clang_opt_flags.append(f"-gsplit-dwarf={args.split_dwarf}")

    # The following args slices manipulating may be confusing. The first 3 element of opt_args are:
    #   1. a spliter "--", it's not used anywhere;
    #   2. the fbcc wrapper script path
    #   3. the "-cc" arg pointing to the compiler we use
    # EXAMPLE: ['--', 'buck-out/v2/gen/fbcode/8e3db19fe005003a/tools/build/buck/wrappers/__fbcc__/fbcc', '--cc=fbcode/third-party-buck/platform010/build/llvm-fb/12/bin/clang++', '--target=x86_64-redhat-linux-gnu', ...]
    clang_cc1_flags = _cleanup_flags(args.opt_args[2:] + clang_opt_flags)
    if clang_cc1_flags is None:
        return EXIT_FAILURE

    fbcc_cmd = args.opt_args[1:3] + clang_cc1_flags
    if args.debug:
        # Print fbcc commandline and exit.
        print(" ".join(fbcc_cmd))
        return EXIT_SUCCESS

    subprocess.check_call(fbcc_cmd)
    if os.stat(args.out).st_size == 0:
        print("error: opt produced empty file")
        return EXIT_FAILURE
    return EXIT_SUCCESS


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import re
import subprocess


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--tar-flag", action="append")
    parser.add_argument("--tar-archive")
    parser.add_argument("--exclude", action="append")
    parser.add_argument("--out")
    args = parser.parse_args()

    exclusions = [re.compile(e) for e in args.exclude]
    files = subprocess.check_output(
        ["tar", "--list", "-f", args.tar_archive] + args.tar_flag, encoding="utf-8"
    )
    files = [f.strip() for f in files.split()]

    with open(args.out, "w", encoding="utf-8") as out:
        for f in files:
            if all(excl.match(f) is None for excl in exclusions):
                continue
            out.write("{}\n".format(f))


if __name__ == "__main__":
    main()



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import unittest

from cxx.dist_lto.tools.dist_lto_opt import _filter_flags


class TestDistLtoOpt(unittest.TestCase):
    def test_filter_flags(self):
        inputs = [
            "-not-opt-flag",
            "-Wl,-plugin-opt,-llvm_flag1",
            "-Wl,-plugin-opt=-llvm_flag1_different_format",
            "-Wl,-plugin-opt,sample-profile=/fbcode/abc/autofdo.profile",
            "-Wl,--lto-whole-program-visibility",
            "-Wl,-mllvm,-llvm-flag2",
            "-Wl,-plugin-opt,O3",
            "-Wl,-plugin-opt,-function-sections",
            "-Wl,-mllvm,-hot-callsite-threshold=12000",
            "-shared",
        ]
        flags = _filter_flags(inputs)
        self.assertListEqual(
            flags,
            [
                "-O2",
                "-fexperimental-new-pass-manager",
                "-ffunction-sections",
                "-fdata-sections",
                "-mllvm",
                "-llvm_flag1",
                "-mllvm",
                "-llvm_flag1_different_format",
                "-fprofile-sample-use=/fbcode/abc/autofdo.profile",
                "-mllvm",
                "-llvm-flag2",
                "-O3",
                "-mllvm",
                "-hot-callsite-threshold=12000",
                "-fPIC",
            ],
        )

    def test_filter_flags_hhvm_case_rev_0f8618f31(self):
        inputs = [
            "--target=x86_64-redhat-linux-gnu",
            "-nostdinc",
            "-resource-dir",
            "fbcode/third-party-buck/platform010/build/llvm-fb/12/lib/clang/stable",
            "-idirafter",
            "fbcode/third-party-buck/platform010/build/llvm-fb/12/lib/clang/stable/include",
            "-idirafter",
            "fbcode/third-party-buck/platform010/build/glibc/include",
            "-idirafter",
            "fbcode/third-party-buck/platform010/build/kernel-headers/include",
            "-Bfbcode/third-party-buck/platform010/build/binutils/x86_64-facebook-linux/bin",
            "--cflag=--target=x86_64-redhat-linux-gnu",
            "--ar=fbcode/third-party-buck/platform010/build/llvm-fb/12/bin/llvm-ar",
            "-Bfbcode/third-party-buck/platform010/build/glibc/lib",
            "-Bfbcode/third-party-buck/platform010/tools/gcc/lib/gcc/x86_64-redhat-linux-gnu/trunk",
            "-Lfbcode/third-party-buck/platform010/build/libgcc/lib/gcc/x86_64-facebook-linux/trunk",
            "-Wl,-nostdlib",
            "-Wl,--dynamic-linker,/usr/local/fbcode/platform010/lib/ld.so",
            "-Wl,--disable-new-dtags",
            "-Bfbcode/third-party-buck/platform010/build/binutils/x86_64-facebook-linux/bin",
            "-Bbuck-out/v2/gen/fbcode/8e3db19fe005003a/third-party-buck/platform010/build/llvm-fb/12/__lld_path__/lld_path/bin",
            "-Wl,--no-mmap-output-file",
            "-nodefaultlibs",
            "--target=x86_64-redhat-linux-gnu",
            "-Lfbcode/third-party-buck/platform010/build/glibc/lib",
            "-Lfbcode/third-party-buck/platform010/build/libgcc/lib",
            "-Wl,-z,notext",
            "-Wl,-z,relro",
            "-Wl,--gc-sections",
            "-fuse-ld=lld",
            "-Wl,--discard-section=.nv_fatbin",
            "-Wl,--discard-section=.nvFatBinSegment",
            "-Wl,--discard-section=.rela.debug_info",
            "-Wl,--discard-section=.rela.debug_ranges",
            "-Wl,--discard-section=.rela.debug_loc",
            "-Wl,--discard-section=.rela.debug_line",
            "-Wl,--discard-section=.rela.debug_aranges",
            "-Wl,--discard-section=.rela.debug_types",
            "-Wl,-O1",
            "-Wl,--build-id=sha1",
            "-Wl,-mllvm,-hot-callsite-threshold=12000",
            "-Wl,--lto-whole-program-visibility",
            "-fwhole-program-vtables",
            "-fexperimental-new-pass-manager",
            "-Wl,--no-discard-section=.nv_fatbin",
            "-Wl,--no-discard-section=.nvFatBinSegment",
            "fbcode/tools/build/move_gpu_sections_implicit_linker_script.txt",
            "-fuse-ld=lld",
            "--build-info=full",
            "--build-info-build-mode=opt-hhvm-lto",
            "--build-info-build-tool=buck2",
            "--build-info-compiler=clang",
            "--build-info-fdo-profile=",
            "--build-info-platform=platform010",
            "--build-info-rule=fbcode:scripts/fnz/minimal:minimal",
            "--build-info-rule-type=cpp_binary",
            "-flto=thin",
            "-Wl,-plugin-opt,-function-sections",
            "-Wl,-plugin-opt,-profile-guided-section-prefix=false",
            "-Wl,-plugin-opt,-generate-type-units",
            "-Wl,-plugin-opt,-enable-lto-ir-verification=false",
            "-Xlinker",
            "--export-dynamic-symbol=mallocx",
            "-Xlinker",
            "--export-dynamic-symbol=rallocx",
            "-Xlinker",
            "--export-dynamic-symbol=xallocx",
            "-Xlinker",
            "--export-dynamic-symbol=sallocx",
            "-Xlinker",
            "--export-dynamic-symbol=dallocx",
            "-Xlinker",
            "--export-dynamic-symbol=sdallocx",
            "-Xlinker",
            "--export-dynamic-symbol=nallocx",
            "-Xlinker",
            "--export-dynamic-symbol=mallctl",
            "-Xlinker",
            "--export-dynamic-symbol=mallctlnametomib",
            "-Xlinker",
            "--export-dynamic-symbol=mallctlbymib",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_stats_print",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_usable_size",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_message",
        ]
        flags = _filter_flags(inputs)
        self.assertIsNotNone(flags)
        self.assertListEqual(
            flags,
            [
                "-O2",
                "-fexperimental-new-pass-manager",
                "-ffunction-sections",
                "-fdata-sections",
                "-mllvm",
                "-hot-callsite-threshold=12000",
                "-mllvm",
                "-profile-guided-section-prefix=false",
                "-mllvm",
                "-generate-type-units",
                "-mllvm",
                "-enable-lto-ir-verification=false",
            ],
        )

    def test_filter_flags_unicorn_case_rev_0f8618f31(self):
        inputs = [
            "--ld=fbcode/third-party-buck/platform010/build/llvm-fb/12/bin/clang++",
            "--cc=buck-out/v2/gen/fbcode/8e3db19fe005003a/tools/build/buck/wrappers/__fbcc__/fbcc --cc=fbcode/third-party-buck/platform010/build/llvm-fb/12/bin/clang --target=x86_64-redhat-linux-gnu -nostdinc -resource-dir fbcode/third-party-buck/platform010/build/llvm-fb/12/lib/clang/stable -idirafter fbcode/third-party-buck/platform010/build/llvm-fb/12/lib/clang/stable/include -idirafter fbcode/third-party-buck/platform010/build/glibc/include -idirafter fbcode/third-party-buck/platform010/build/kernel-headers/include -Bfbcode/third-party-buck/platform010/build/binutils/x86_64-facebook-linux/bin",
            "--cflag=--target=x86_64-redhat-linux-gnu",
            "--ar=fbcode/third-party-buck/platform010/build/llvm-fb/12/bin/llvm-ar",
            "-Bfbcode/third-party-buck/platform010/build/glibc/lib",
            "-Bfbcode/third-party-buck/platform010/tools/gcc/lib/gcc/x86_64-redhat-linux-gnu/trunk",
            "-Lfbcode/third-party-buck/platform010/build/libgcc/lib/gcc/x86_64-facebook-linux/trunk",
            "-Wl,-nostdlib",
            "-Wl,--dynamic-linker,/usr/local/fbcode/platform010/lib/ld.so",
            "-Wl,--disable-new-dtags",
            "-Bfbcode/third-party-buck/platform010/build/binutils/x86_64-facebook-linux/bin",
            "-Bbuck-out/v2/gen/fbcode/8e3db19fe005003a/third-party-buck/platform010/build/llvm-fb/12/__lld_path__/lld_path/bin",
            "-Wl,--no-mmap-output-file",
            "-nodefaultlibs",
            "--target=x86_64-redhat-linux-gnu",
            "-Lfbcode/third-party-buck/platform010/build/glibc/lib",
            "-Lfbcode/third-party-buck/platform010/build/libgcc/lib",
            "-Wl,-z,notext",
            "-Wl,-z,relro",
            "-Wl,--gc-sections",
            "-fuse-ld=lld",
            "-Wl,--discard-section=.nv_fatbin",
            "-Wl,--discard-section=.nvFatBinSegment",
            "-Wl,--discard-section=.rela.debug_info",
            "-Wl,--discard-section=.rela.debug_ranges",
            "-Wl,--discard-section=.rela.debug_loc",
            "-Wl,--discard-section=.rela.debug_line",
            "-Wl,--discard-section=.rela.debug_aranges",
            "-Wl,--discard-section=.rela.debug_types",
            "-Wl,-O1",
            "-Wl,--build-id=sha1",
            "-fexperimental-new-pass-manager",
            "-Xlinker",
            "-znow",
            "-Xlinker",
            "--emit-relocs",
            "--build-info=full",
            "--build-info-build-mode=opt-clang-thinlto",
            "--build-info-build-tool=buck2",
            "--build-info-compiler=clang",
            "--build-info-fdo-profile=fbcode//fdo/autofdo/default_profile:autofdo",
            "--build-info-platform=platform010",
            "--build-info-rule=fbcode:unicorn:index_server",
            "--build-info-rule-type=cpp_binary",
            "-flto=thin",
            "-Wl,-plugin-opt,sample-profile=buck-out/v2/gen/fbcode/40fc99293b37c503/fdo/autofdo/default_profile/__autofdo__/out/profile",
            "-Wl,-plugin-opt,-function-sections",
            "-Wl,-plugin-opt,-profile-guided-section-prefix=false",
            "-Wl,-plugin-opt,-generate-type-units",
            "-Wl,-plugin-opt,-enable-lto-ir-verification=false",
            "-Xlinker",
            "--push-state",
            "-Xlinker",
            "--no-as-needed",
            "-Xlinker",
            "--pop-state",
            "-Wl,--undefined,Global",
            "-Wl,--undefined,Local",
            "-Wl,--undefined,MockConnection",
            "-Wl,--undefined,Global",
            "-Wl,--undefined,Local",
            "-Wl,--undefined,MockConnection",
            "-Xlinker",
            "--start-group",
            "fbcode/third-party-buck/platform010/build/IntelComposerXE/mkl/lib/intel64/libmkl_intel_lp64.a",
            "fbcode/third-party-buck/platform010/build/IntelComposerXE/mkl/lib/intel64/libmkl_core.a",
            "fbcode/third-party-buck/platform010/build/IntelComposerXE/mkl/lib/intel64/libmkl_intel_thread.a",
            "-Xlinker",
            "--end-group",
            "-lpthread",
            "-Xlinker",
            "--export-dynamic-symbol=mallocx",
            "-Xlinker",
            "--export-dynamic-symbol=rallocx",
            "-Xlinker",
            "--export-dynamic-symbol=xallocx",
            "-Xlinker",
            "--export-dynamic-symbol=sallocx",
            "-Xlinker",
            "--export-dynamic-symbol=dallocx",
            "-Xlinker",
            "--export-dynamic-symbol=sdallocx",
            "-Xlinker",
            "--export-dynamic-symbol=nallocx",
            "-Xlinker",
            "--export-dynamic-symbol=mallctl",
            "-Xlinker",
            "--export-dynamic-symbol=mallctlnametomib",
            "-Xlinker",
            "--export-dynamic-symbol=mallctlbymib",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_stats_print",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_usable_size",
            "-Xlinker",
            "--export-dynamic-symbol=malloc_message",
        ]
        flags = _filter_flags(inputs)
        self.assertIsNotNone(flags)
        self.assertListEqual(
            flags,
            [
                "-O2",
                "-fexperimental-new-pass-manager",
                "-ffunction-sections",
                "-fdata-sections",
                "-fprofile-sample-use=buck-out/v2/gen/fbcode/40fc99293b37c503/fdo/autofdo/default_profile/__autofdo__/out/profile",
                "-mllvm",
                "-profile-guided-section-prefix=false",
                "-mllvm",
                "-generate-type-units",
                "-mllvm",
                "-enable-lto-ir-verification=false",
            ],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Creates a Python Source DB JSON file from Python manifest JSON file (e.g. for use with Pyre).

Sources and dependencies are passed in via source manifest files, which are
merged by this script:

$ ./make_source_db_no_deps.py \
      my_rule.manifest.json \
      --output db_no_deps.json

The output format of the source DB is:

{
  <source1-name>: <source1-path>,
  <source2-name>: <source2-path>,
  ...
}
"""

import argparse
import json
import sys


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def _load(path):
    with open(path) as f:
        return json.load(f)


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("sources")
    args = parser.parse_args(argv[1:])

    sources = {name: path for name, path, _ in _load(args.sources)}
    json.dump(sources, args.output, indent=2)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import os
import re
from io import TextIOWrapper
from typing import Dict, Iterable, List


class Module:
    def __init__(self, name: str) -> None:
        self.name: str = name
        self.headers: List[str] = []
        self.submodules: Dict[str, Module] = {}

    def add_header(self, src: str) -> None:
        self.headers.append(src)

    def get_submodule(self, name: str) -> "Module":
        if name not in self.submodules:
            self.submodules[name] = Module(name)

        return self.submodules[name]

    def render(self, f: TextIOWrapper, path_prefix: str, indent: int = 0) -> None:
        space = " " * indent
        f.write(f"{space}module {self.name} {{\n")

        submodule_names = set()
        for submodule_name in sorted(self.submodules.keys()):
            submodule = self.submodules[submodule_name]

            # remove any extensions for readibility
            sanitized_name = os.path.splitext(submodule_name)[0]

            # module names can only be ascii or _
            sanitized_name = re.sub(r"[^A-Za-z0-9_]", "_", sanitized_name)
            if sanitized_name[0].isdigit():
                sanitized_name = "_" + sanitized_name

            # avoid any collisions with other files
            while sanitized_name in submodule_names:
                sanitized_name += "_"

            submodule_names.add(sanitized_name)
            submodule.name = sanitized_name
            submodule.render(f, path_prefix, indent + 4)

        header_space = " " * (indent + 4)
        for h in sorted(self.headers):
            f.write(f'{header_space}header "{os.path.join(path_prefix, h)}"\n')

        if self.headers:
            f.write(f"{header_space}export *\n")

        f.write(f"{space}}}\n")


def _write_single_module(
    f: TextIOWrapper, name: str, headers: Iterable[str], path_prefix: str
) -> None:
    module = Module(name)
    for h in headers:
        module.add_header(h)

    module.render(f, path_prefix)


def _write_submodules(
    f: TextIOWrapper, name: str, headers: Iterable[str], path_prefix: str
) -> None:
    # Create a tree of nested modules, one for each path component.
    root_module = Module(name)
    for h in headers:
        module = root_module
        for i, component in enumerate(h.split(os.sep)):
            if i == 0 and component == name:
                # The common case is we have a singe header path prefix that matches the module name.
                # In this case we add the headers directly to the root module.
                pass
            else:
                module = module.get_submodule(component)

        module.add_header(h)

    root_module.render(f, path_prefix)


def _write_swift_header(f: TextIOWrapper, name: str, swift_header_path: str) -> None:
    f.write(
        f"""
module {name}.Swift {{
    header "{swift_header_path}"
    requires objc
}}
"""
    )


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output", required=True, help="The path to write the modulemap to"
    )
    parser.add_argument("--name", required=True, help="The name of the module")
    parser.add_argument(
        "--swift-header", help="If this is a mixed module extend with this Swift header"
    )
    parser.add_argument(
        "--use-submodules",
        action="store_true",
        help="If set produce a modulemap with per-header submodules",
    )
    parser.add_argument(
        "--symlink-tree",
        required=True,
    )
    parser.add_argument(
        "mappings", nargs="*", default=[], help="A list of import paths"
    )
    args = parser.parse_args()

    output_dir = os.path.dirname(args.output)
    path_prefix = os.path.relpath(args.symlink_tree, output_dir)

    with open(args.output, "w") as f:
        if args.use_submodules:
            # pyre-fixme[6]: For 4th param expected `str` but got `bytes`.
            _write_submodules(f, args.name, args.mappings, path_prefix)
        else:
            # pyre-fixme[6]: For 4th param expected `str` but got `bytes`.
            _write_single_module(f, args.name, args.mappings, path_prefix)

        if args.swift_header:
            swift_header_name = os.path.relpath(args.swift_header, output_dir)
            _write_swift_header(
                f,
                args.name,
                os.path.join(
                    "swift-extended_symlink_tree",
                    args.name,
                    str(swift_header_name),
                ),
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import sys
from typing import List


def main(argv: List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("--extension", action="append", default=[])
    args = parser.parse_args(argv[1:])
    out_file = args.output

    externs = []
    table = [
        "struct _inittab _static_extension_info[] = {",
    ]
    for python_name in args.extension:
        module_name, pyinit_func = python_name.split(":")
        # If this is a top level module we do not suffix the PyInit_ symbol
        externs.append(f"PyMODINIT_FUNC {pyinit_func}(void);")
        table.append(f'  {{ "{module_name}", {pyinit_func} }},')
    table.append("  { nullptr, nullptr },")
    table.append("};")

    out_lines = (
        [
            '#include "Python.h"',
            '#include "import.h"',
        ]
        + externs
        + table
    )

    for line in out_lines:
        print(line, file=out_file)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

# pyre-unsafe

import argparse
import os
import subprocess
import sys
from pathlib import Path


def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--cgo", action="append", default=[])
    parser.add_argument("--output", required=True, type=Path)
    parser.add_argument("--cpp", action="append", default=[])
    parser.add_argument("srcs", type=Path, nargs="*")
    args = parser.parse_args(argv[1:])

    output = args.output.resolve(strict=False)
    os.makedirs(output, exist_ok=True)

    os.environ["CC"] = args.cpp[0]

    cmd = []
    cmd.extend(args.cgo)
    # cmd.append("-importpath={}")
    # cmd.append("-srcdir={}")
    cmd.append(f"-objdir={output}")
    # cmd.append(cgoCompilerFlags)
    cmd.append("--")
    # cmd.append(cxxCompilerFlags)
    cmd.extend(args.cpp[1:])
    cmd.extend(args.srcs)
    subprocess.check_call(cmd)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Create the link tree for inplace Python binaries.

This does a few things:
    - Allows remapping of source files (via the srcs/dests arguments) and resources.
    - Merges extracted .whl files into the tree
    - Adds __init__.py where needed
    - Writes out a bootstrapper pex script that knows where this symlink tree is,
      and uses it, along with the provided entry point to run the python script.
      It does this by replacing a few special strings like <MODULES_DIR> and
      <MAIN_MODULE>

A full usage might be something like this:

$ cat srcs
srcs/foo.py
srcs/bar.py
third-party/baz.whl__extracted
$ cat dests
lib/foo.py
bar.py
.
$ ls third-party/baz.whl__extracted
baz/tp_foo.py
baz/tp_bar.py
$ cat template.in
(see prelude/python/run_inplace_lite.py.in)
$ ./make_pex_inplace.py  \\
    --template prelude/python/run_inplace.py.in \\
    --module-srcs=@srcs \\
    --module-dests=@dests \\
    # This is the symlink tree \\
    --modules-dir=bin__link-tree
$ find bin__link-tree
lib/__init__.py
lib/foo.py
bar.py
baz/tp_foo.py
baz/tp_bar.py
"""

import argparse
import errno
import json
import os
from pathlib import Path
from typing import Dict, Set, Tuple

# Suffixes which should trigger `__init__.py` additions.
# TODO(agallaher): This was coped from v1, but some things below probably
# don't need to be here (e.g. `.pyd`).
_MODULE_SUFFIXES = {
    ".dll",
    ".py",
    ".pyd",
    ".so",
}


def parse_args() -> argparse.Namespace:
    # TODO(nmj): Go back and verify all of the various flags that make_xar
    #                 takes, and standardize on that so that the calling convention
    #                 is the same regardless of the "make_X" binary that's used.
    parser = argparse.ArgumentParser(
        description=(
            "Create a python inplace binary, writing a symlink tree to a directory, "
            "and a bootstrapper pex file to file"
        ),
        fromfile_prefix_chars="@",
    )
    parser.add_argument(
        "--module-manifest",
        action="append",
        dest="module_manifests",
        default=[],
        help="A path to a JSON file with modules to be linked.",
    )
    parser.add_argument(
        "--resource-manifest",
        action="append",
        dest="resource_manifests",
        default=[],
        help="A path to a JSON file with resources to be linked.",
    )
    parser.add_argument(
        "--native-library-src",
        type=Path,
        dest="native_library_srcs",
        action="append",
        default=[],
        help="A list of native libraries to use",
    )
    parser.add_argument(
        "--native-library-dest",
        type=Path,
        dest="native_library_dests",
        action="append",
        default=[],
        help=(
            "A list of relative destination paths for each of the native "
            "libraries in --native-library-src"
        ),
    )
    parser.add_argument(
        "--dwp-src",
        type=Path,
        dest="dwp_srcs",
        action="append",
        default=[],
        help="A list of dwp for native libraries to use",
    )
    parser.add_argument(
        "--dwp-dest",
        type=Path,
        dest="dwp_dests",
        action="append",
        default=[],
        help=(
            "A list of relative destination paths for each of the dwp for native "
            "libraries in --dwp-src"
        ),
    )
    parser.add_argument(
        "--native-library-manifest",
        action="append",
        dest="native_library_manifests",
        default=[],
        help="A path to a JSON file with native libraries to be linked.",
    )
    parser.add_argument(
        "--modules-dir",
        required=True,
        type=Path,
        help="The link tree directory to write to",
    )

    return parser.parse_args()


def add_path_mapping(
    path_mapping: Dict[Path, Tuple[str, str]],
    dirs_to_create: Set[Path],
    src: Path,
    new_dest: Path,
    origin: str = "unknown",
) -> None:
    """
    Add the mapping of a destination path into `path_mapping`, by getting the
    relative path to the source, and making sure that there are no
    collisions (and erroring in that case)
    """

    def format_src(src: str, origin: str) -> str:
        out = "`{}`".format(src)
        if origin is not None:
            out += " (from {})".format(origin)
        return out

    link_path = os.path.relpath(src, new_dest.parent)
    if new_dest in path_mapping:
        prev, prev_origin = path_mapping[new_dest]
        if prev != link_path:
            raise ValueError(
                "Destination path `{}` specified at both {} and {} (`{}` before relativisation)".format(
                    new_dest,
                    format_src(link_path, origin),
                    format_src(prev, prev_origin),
                    src,
                )
            )
    path_mapping[new_dest] = (link_path, origin)
    dirs_to_create.add(new_dest.parent)


def _lexists(path: Path) -> bool:
    """
    Like `Path.exists()` but works on dangling. symlinks
    """

    try:
        path.lstat()
    except FileNotFoundError:
        return False
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False
        raise
    return True


def create_modules_dir(args: argparse.Namespace) -> None:
    args.modules_dir.mkdir(parents=True, exist_ok=True)

    # Mapping of destination files -> the symlink target (e.g. "../foo")
    path_mapping: Dict[Path, Tuple[str, str]] = {}
    # Set of directories that need to be created in the link tree before
    # symlinking
    dirs_to_create: Set[Path] = set()
    # Set of __init__.py files that need to be created at the end of the
    # link tree building if they don't exist, so that python recognizes them
    # as modules.
    init_py_paths = set()

    # Link entries from manifests.
    for manifest in args.module_manifests:
        with open(manifest) as manifest_file:
            for dest, src, origin in json.load(manifest_file):
                dest = Path(dest)

                # Add `__init__.py` files for all parent dirs (except the root).
                if dest.suffix in _MODULE_SUFFIXES:
                    package = dest.parent
                    while package != Path("") and package not in init_py_paths:
                        init_py_paths.add(package)
                        package = package.parent

                add_path_mapping(
                    path_mapping,
                    dirs_to_create,
                    src,
                    args.modules_dir / dest,
                    origin=origin,
                )

    for manifest in args.resource_manifests + args.native_library_manifests:
        with open(manifest) as manifest_file:
            for dest, src, origin in json.load(manifest_file):
                add_path_mapping(
                    path_mapping,
                    dirs_to_create,
                    src,
                    args.modules_dir / dest,
                    origin=origin,
                )

    if args.native_library_srcs:
        for src, dest in zip(args.native_library_srcs, args.native_library_dests):
            new_dest = args.modules_dir / dest
            add_path_mapping(path_mapping, dirs_to_create, src, new_dest)

    if args.dwp_srcs:
        for src, dest in zip(args.dwp_srcs, args.dwp_dests):
            new_dest = args.modules_dir / dest
            add_path_mapping(path_mapping, dirs_to_create, src, new_dest)

    for d in dirs_to_create:
        d.mkdir(parents=True, exist_ok=True)

    for dest, (target, _origin) in path_mapping.items():
        try:
            os.symlink(target, dest)
        except OSError:
            if _lexists(dest):
                if os.path.islink(dest):
                    raise ValueError(
                        "{} already exists, and is linked to {}. Cannot link to {}".format(
                            dest, os.readlink(dest), target
                        )
                    )
                else:
                    raise ValueError(
                        "{} already exists. Cannot link to {}".format(dest, target)
                    )
            else:
                raise

    # Fill in __init__.py for sources that were provided by the user
    # These are filtered such that we only create this for sources specified
    # by the user; if a .whl fortgets an __init__.py file, that's their problem
    for init_py_dir in init_py_paths:
        init_py_path = args.modules_dir / init_py_dir / "__init__.py"
        # We still do this check because python insists on touching some read only
        # files and blows up somtimes.
        if not _lexists(init_py_path):
            init_py_path.touch(exist_ok=True)


def main() -> None:
    args = parse_args()
    create_modules_dir(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Utility to create compilation DBs

$ make_comp_db.py gen --output=entry.json foo.cpp -- g++ -c -fPIC
$ make_comp_db.py gen --output=entry2.json foo2.cpp -- g++ -c -fPIC
$ make_comp_db.py merge --output=comp_db.json entry.json entry2.json
"""

# pyre-unsafe

import argparse
import json
import shlex
import sys


def gen(args):
    """
    Generate a single compilation command in JSON form.
    """

    entry = {}
    entry["file"] = args.directory + "/" + args.filename
    entry["directory"] = "."

    arguments = []
    for arg in args.arguments:
        if arg.startswith("@"):
            with open(arg[1:]) as argsfile:
                for line in argsfile:
                    # The argsfile's arguments are separated by newlines; we
                    # don't want those included in the argument list.
                    arguments.append(" ".join(shlex.split(line)))
        else:
            arguments.append(arg)
    entry["arguments"] = arguments

    json.dump(entry, args.output, indent=2)


def merge(args):
    """
    Merge multiple compilation DB commands into a single DB.
    """

    entries = []
    for entry in args.entries:
        with open(entry) as f:
            entries.append(json.load(f))

    json.dump(entries, args.output, indent=2)


def main(argv):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    parser_gen = subparsers.add_parser("gen")
    parser_gen.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser_gen.add_argument("filename")
    parser_gen.add_argument("directory")
    parser_gen.add_argument("arguments", nargs="*")
    parser_gen.set_defaults(func=gen)

    parser_merge = subparsers.add_parser("merge")
    parser_merge.add_argument(
        "--output", type=argparse.FileType("w"), default=sys.stdout
    )
    parser_merge.add_argument("entries", nargs="*")
    parser_merge.set_defaults(func=merge)

    args = parser.parse_args(argv[1:])
    args.func(args)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Run on a directory of Go source files and print out a list of srcs that should
be compiled.

Example:

 $ ./filter_srcs.py --output srcs.txt src/dir/

"""

# pyre-unsafe

import argparse
import json
import os
import subprocess
import sys
from pathlib import Path


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--go", default="go", type=Path)
    parser.add_argument("--tests", action="store_true")
    parser.add_argument("--tags", default="")
    parser.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("srcdir", type=Path)
    args = parser.parse_args(argv[1:])

    # Find all source sub-dirs, which we'll need to run `go list` from.
    roots = set()
    for root, _dirs, _files in os.walk(args.srcdir):
        roots.add(root)

    # Run `go list` on all source dirs to filter input sources by build pragmas.
    for root in roots:
        out = subprocess.check_output(
            [
                "env",
                "-i",
                "GOARCH={}".format(os.environ.get("GOARCH", "")),
                "GOOS={}".format(os.environ.get("GOOS", "")),
                "CGO_ENABLED={}".format(os.environ.get("CGO_ENABLED", "0")),
                "GO111MODULE=off",
                "GOCACHE=/tmp",
                args.go.resolve(),
                "list",
                "-e",
                "-json",
                "-tags",
                args.tags,
                ".",
            ],
            cwd=root,
        ).decode("utf-8")

        # Parse JSON output and print out sources.
        idx = 0
        decoder = json.JSONDecoder()
        while idx < len(out) - 1:
            # The raw_decode method fails if there are any leading spaces, e.g. " {}" fails
            # so manually trim the prefix of the string
            if out[idx].isspace():
                idx += 1
                continue

            obj, idx = decoder.raw_decode(out, idx)
            if args.tests:
                types = ["GoFiles", "TestGoFiles", "XTestGoFiles"]
            else:
                types = ["GoFiles", "SFiles"]
            for typ in types:
                for src in obj.get(typ, []):
                    src = Path(obj["Dir"]) / src
                    src = src.resolve().relative_to(os.getcwd())
                    print(src, file=args.output)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import shutil
import sys
from typing import List


def main(argv: List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--to")
    parser.add_argument("--from", dest="from_")
    args = parser.parse_args(argv[1:])
    shutil.copy(args.from_, args.to)
    return 0


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
A simple wrapper around a distributed thinlto index command to fit into buck2's
distributed thinlto build.

This reads in a couple of things:
    1. The "meta" file. This is a list of tuples of (object file, index output,
    plan output). All items are line-separated (so each tuple is three lines).
    2. The index and link plan output paths
    3. The commands for the actual index command.

It will invoke the index command and then copy the index outputs to the
requested locations and write a plan for each of those objects. This "plan" is
a simple json file with the most important thing being a list of the indices
of the imports needed for that file.

It will then additionally write a link plan, which is just a translation of
the thinlto index (which lists the objects actually needed for the final link).


Both opt and link plans use indices to refer to other files because it allows the bzl
code to easily map back to other objects held in buck memory.
"""

# pyre-unsafe

import argparse
import json
import os
import os.path
import subprocess
import sys
from typing import List


def _get_argsfile(args) -> str:
    # go through the flags passed to linker and find the index argsfile
    argsfiles = list(
        filter(lambda arg: arg.endswith("thinlto.index.argsfile"), args.index_args)
    )
    assert (
        len(argsfiles) == 1
    ), f"expect only 1 argsfile but seeing multiple ones: {argsfiles}"
    argsfile = argsfiles[0]
    if argsfile.startswith("@"):
        argsfile = argsfile[1:]
    return argsfile


def _extract_lib_search_path(argsfile_path: str) -> List[str]:
    lib_search_path = []
    with open(argsfile_path) as argsfile:
        for line in argsfile:
            if line.startswith("-L"):
                lib_search_path.append(line.strip())
    return lib_search_path


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--meta")
    parser.add_argument("--index")
    parser.add_argument("--link-plan")
    parser.add_argument("--final-link-index")
    parser.add_argument("index_args", nargs=argparse.REMAINDER)
    args = parser.parse_args(argv[1:])

    subprocess.check_call(args.index_args[1:])

    bitcode_suffix = ".thinlto.bc"
    imports_suffix = ".imports"
    opt_objects_suffix = ".opt.o"  # please note the files are not exist yet, this is to generate the index file use in final link

    with open(args.meta) as meta:
        meta_lines = [line.strip() for line in meta.readlines()]

    def read_imports(path, imports_path):
        with open(imports_path) as infile:
            return [line.strip() for line in infile.readlines()]

    def index_path(path):
        return os.path.join(args.index, path)

    # The meta file comes directly from dist_lto.bzl and consists of a list of
    # 7-tuples of information. It is easiest for us to write each tuple member
    # as a separate line in Starlark, so these 7-tuples are encoded in groups
    # of seven lines.
    #
    # The seven pieces of information are:
    #  1. The path to the source bitcode file. This is used as an index into
    #     a dictionary (`mapping`) that records much of the metadata coming
    #     from these lines.
    #  2. The path to an output bitcode file. This script is expected to place a
    #     ThinLTO index file at this location (suffixed `.thinlto.bc`).
    #  3. The path to an output plan. This script is expected to place a link
    #     plan here (a JSON document indicating which other object files this)
    #     object file depends on, among other things.
    #  4. The link data's index in the Starlark array.
    #  5. If this object file came from an archive, the name of the archive. Otherwise,
    #     this line is empty.
    #  6. If this object file came from an archive, the path to an output plan.
    #     This script is expected to produce an archive link plan here (a JSON)
    #     document similar to the object link plan, except containing link
    #     information for every file in the archive from which this object
    #     came. Otherwise, this line is empty.
    #  7. If this object file came from an archive, the indexes directory of that
    #     archive. This script is expected to place all ThinLTO indexes derived
    #     from object files originating from this archive in that directory.
    #     Otherwise, this line is empty.
    #
    # There are two indices that are derived from this meta file: the object
    # index (mapping["index"]) and the archive index (mapping["archive_index"]).
    # These indices are indices into Starlark arrays for all objects and archive
    # linkables, respectively. This script does not inspect them.
    mapping = {}
    archives = {}
    for i in range(0, len(meta_lines), 7):
        path = meta_lines[i]
        output = meta_lines[i + 1]
        plan_output = meta_lines[i + 2]
        idx = int(meta_lines[i + 3])
        archive_name = meta_lines[i + 4]
        archive_plan = meta_lines[i + 5]
        archive_index_dir = meta_lines[i + 6]

        archive_idx = idx if output == "" else None  # archives do not have outputs
        mapping[path] = {
            "output": output,
            "plan_output": plan_output,
            "index": idx,
            "archive_index": archive_idx,
            "archive_name": archive_name,
        }
        if archive_idx is not None:
            archives[idx] = {
                "name": archive_name,
                "objects": [],
                "plan": archive_plan,
                "index_dir": archive_index_dir,
            }

    non_lto_objects = {}
    for path, data in sorted(mapping.items(), key=lambda v: v[0]):
        output_loc = data["output"]
        if os.path.exists(output_loc):
            continue

        if data["archive_index"] is not None:
            archives[data["archive_index"]]["objects"].append(path)
            continue

        bc_file = index_path(path) + bitcode_suffix
        imports_path = index_path(path) + imports_suffix
        os.makedirs(os.path.dirname(output_loc), exist_ok=True)

        if os.path.exists(imports_path):
            assert os.path.exists(bc_file), "missing bc file for %s" % path
            os.rename(bc_file, output_loc)
            imports = read_imports(path, imports_path)
            imports_list = []
            archives_list = []
            for path in imports:
                entry = mapping[path]
                if entry["archive_index"] is not None:
                    archives_list.append(int(entry["archive_index"]))
                else:
                    imports_list.append(entry["index"])
            plan = {
                "imports": imports_list,
                "archive_imports": archives_list,
                "index": data["index"],
                "bitcode_file": bc_file,
                "path": path,
                "is_bc": True,
            }
        else:
            non_lto_objects[data["index"]] = 1
            with open(output_loc, "w"):
                pass
            plan = {
                "is_bc": False,
            }

        with open(data["plan_output"], "w") as planout:
            json.dump(plan, planout, sort_keys=True)

    for archive in archives.values():
        # For archives, we must produce a plan that provides Starlark enough
        # information about how to launch a dynamic opt for each object file
        # in the archive.
        archive_plan = {}

        # This is convenient to store, since it's difficult for Starlark to
        # calculate it.
        archive_plan["base_dir"] = os.path.dirname(archive["plan"])
        object_plans = []
        for obj in archive["objects"]:
            imports_path = index_path(obj) + imports_suffix
            output_path = archive["index_dir"]
            os.makedirs(output_path, exist_ok=True)
            if os.path.exists(imports_path):
                bc_file = index_path(obj) + bitcode_suffix
                os.rename(bc_file, os.path.join(output_path, os.path.basename(bc_file)))
                imports = read_imports(path, imports_path)

                imports_list = []
                archives_list = []
                for path in imports:
                    entry = mapping[path]
                    if entry["archive_index"] is not None:
                        archives_list.append(int(entry["archive_index"]))
                    else:
                        imports_list.append(entry["index"])
                object_plans.append(
                    {
                        "is_bc": True,
                        "path": obj,
                        "imports": imports_list,
                        "archive_imports": archives_list,
                        "bitcode_file": os.path.join(
                            output_path, os.path.basename(bc_file)
                        ),
                    }
                )
            else:
                object_plans.append(
                    {
                        "is_bc": False,
                        "path": obj,
                    }
                )

        archive_plan["objects"] = object_plans
        with open(archive["plan"], "w") as planout:
            json.dump(archive_plan, planout, sort_keys=True)

    # We read the `index`` and `index.full`` files produced by linker in index stage
    # and translate them to 2 outputs:
    # 1. A link plan build final_link args. (This one may be able to be removed if we refactor the workflow)
    # 2. A files list (*.final_link_index) used for final link stage which includes all the
    #    files needed. it's based on index.full with some modification, like path updates
    #    and redundent(added by toolchain) dependencies removing.
    index = {}
    index_files_set = set()
    # TODO(T130322878): since we call linker wrapper twice (in index and in final_link), to avoid these libs get
    # added twice we remove them from the index file for now.
    KNOWN_REMOVABLE_DEPS_SUFFIX = [
        "glibc/lib/crt1.o",
        "glibc/lib/crti.o",
        "crtbegin.o",
        "crtbeginS.o",
        ".build_info.o",
        "crtend.o",
        "crtendS.o",
        "glibc/lib/crtn.o",
    ]
    with open(index_path("index")) as indexfile:
        for line in indexfile:
            line = line.strip()
            index_files_set.add(line)
            path = os.path.relpath(line, start=args.index)
            index[mapping[path]["index"]] = 1

    with open(args.link_plan, "w") as outfile:
        json.dump(
            {
                "non_lto_objects": non_lto_objects,
                "index": index,
            },
            outfile,
            indent=2,
            sort_keys=True,
        )

    # Append all search path flags (e.g -Lfbcode/third-party-buck/platform010/build/glibc/lib) from argsfile to final_index
    # this workaround is to make dist_lto compatible with link_group. see T136415235 for more info
    argsfile = _get_argsfile(args)
    lib_search_path = _extract_lib_search_path(argsfile)

    # build index file for final link use
    with open(index_path("index.full")) as full_index_input, open(
        args.final_link_index, "w"
    ) as final_link_index_output:
        final_link_index_output.write("\n".join(lib_search_path) + "\n")
        for line in full_index_input:
            line = line.strip()
            if any(filter(line.endswith, KNOWN_REMOVABLE_DEPS_SUFFIX)):
                continue
            path = os.path.relpath(line, start=args.index)
            if line in index_files_set:
                if mapping[path]["output"]:
                    # handle files that were not extracted from archives
                    output = mapping[path]["output"].replace(
                        bitcode_suffix, opt_objects_suffix
                    )
                    final_link_index_output.write(output + "\n")
                elif os.path.exists(index_path(path) + imports_suffix):
                    # handle files built from source that were extracted from archives
                    opt_objects_path = path.replace(
                        "/objects/", "/opt_objects/objects/"
                    )
                    final_link_index_output.write(opt_objects_path + "\n")
                else:
                    # handle pre-built archives
                    final_link_index_output.write(line + "\n")
            else:
                # handle input files that did not come from linker input, e.g. linkerscirpts
                final_link_index_output.write(line + "\n")


sys.exit(main(sys.argv))



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import itertools
import json
import os
from typing import Dict, List, Tuple, TypedDict

# Example VFS overlay in JSON format
# ----------------------------------
# {
#   'version': 0,
#   'roots': [
#     { 'name': 'OUT_DIR', 'type': 'directory',
#       'contents': [
#         { 'name': 'module.map', 'type': 'file',
#           'external-contents': 'INPUT_DIR/actual_module2.map'
#         }
#       ]
#     }
#   ]
# }


class OverlayRoot(TypedDict):
    name: str
    type: str
    contents: List[Dict[str, str]]


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output", required=True, help="The path to write the VFS overlay to"
    )
    parser.add_argument(
        "mappings", nargs="*", default=[], help="A list of virtual paths to real paths"
    )
    args = parser.parse_args()

    if len(args.mappings) % 2 != 0:
        parser.error("mappings must be dest-source pairs")

    # Group the mappings by containing directory
    mappings: Dict[str, List[Tuple[str, str]]] = {}
    for src, dst in itertools.zip_longest(*([iter(args.mappings)] * 2)):
        folder, basename = os.path.split(src)
        mappings.setdefault(folder, []).append((basename, dst))

    with open(args.output, "w") as f:
        json.dump(
            {
                "version": 0,
                "roots": _get_roots(mappings),
            },
            f,
            sort_keys=True,
            indent=4,
        )
        f.write("\n")
        f.flush()


def _get_roots(mappings: Dict[str, List[Tuple[str, str]]]) -> List[OverlayRoot]:
    roots = []
    for folder, file_maps in mappings.items():
        contents = []
        for src, dst in file_maps:
            contents.append(
                {
                    "name": src,
                    "type": "file",
                    "external-contents": dst,
                }
            )

        roots.append(
            {
                "name": folder,
                "type": "directory",
                "contents": contents,
            }
        )

    return roots


if __name__ == "__main__":
    main()



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import json
import os
import re
import sys
from typing import Dict, Iterable, TextIO

# Out-of-date? Update with this command:
#
# xcode-select --print-path | xargs printf '%s/Platforms/iPhoneSimulator.platform/Developer/SDKs/iPhoneSimulator.sdk/System/Library/Frameworks/' | xargs ls | rg '^([A-Z].+)\.framework$' -r '${1}' | xargs printf '    "%s",\n' && xcode-select --print-path | xargs printf '%s/Platforms/iPhoneSimulator.platform/Developer/SDKs/iPhoneSimulator.sdk/usr/include/module.modulemap' | xargs cat | rg '^module ([a-zA-Z0-9_]*) .*$' -r '${1}'| xargs printf '    "%s",\n'
APPLE_SYSTEM_MODULES = {
    "ARKit",
    "AVFAudio",
    "AVFoundation",
    "AVKit",
    "Accelerate",
    "Accessibility",
    "Accounts",
    "AdServices",
    "AdSupport",
    "AddressBook",
    "AddressBookUI",
    "AppClip",
    "AppTrackingTransparency",
    "AssetsLibrary",
    "AudioToolbox",
    "AudioUnit",
    "AuthenticationServices",
    "AutomaticAssessmentConfiguration",
    "BackgroundTasks",
    "BusinessChat",
    "CFNetwork",
    "CallKit",
    "CarPlay",
    "ClassKit",
    "ClockKit",
    "CloudKit",
    "Combine",
    "Contacts",
    "ContactsUI",
    "CoreAudio",
    "CoreAudioKit",
    "CoreAudioTypes",
    "CoreBluetooth",
    "CoreData",
    "CoreFoundation",
    "CoreGraphics",
    "CoreHaptics",
    "CoreImage",
    "CoreLocation",
    "CoreLocationUI",
    "CoreMIDI",
    "CoreML",
    "CoreMedia",
    "CoreMotion",
    "CoreNFC",
    "CoreServices",
    "CoreSpotlight",
    "CoreTelephony",
    "CoreText",
    "CoreVideo",
    "CryptoKit",
    "CryptoTokenKit",
    "DataDetection",
    "DeveloperToolsSupport",
    "DeviceActivity",
    "DeviceCheck",
    "EventKit",
    "EventKitUI",
    "ExposureNotification",
    "ExternalAccessory",
    "FamilyControls",
    "FileProvider",
    "FileProviderUI",
    "Foundation",
    "GLKit",
    "GSS",
    "GameController",
    "GameKit",
    "GameplayKit",
    "GroupActivities",
    "HealthKit",
    "HealthKitUI",
    "HomeKit",
    "IOKit",
    "IOSurface",
    "IdentityLookup",
    "IdentityLookupUI",
    "ImageCaptureCore",
    "ImageIO",
    "Intents",
    "IntentsUI",
    "JavaScriptCore",
    "LinkPresentation",
    "LocalAuthentication",
    "ManagedSettings",
    "ManagedSettingsUI",
    "MapKit",
    "MediaAccessibility",
    "MediaPlayer",
    "MediaToolbox",
    "MessageUI",
    "Messages",
    "Metal",
    "MetalKit",
    "MetalPerformanceShaders",
    "MetalPerformanceShadersGraph",
    "MetricKit",
    "MobileCoreServices",
    "ModelIO",
    "MultipeerConnectivity",
    "MusicKit",
    "NaturalLanguage",
    "NearbyInteraction",
    "Network",
    "NetworkExtension",
    "NewsstandKit",
    "NotificationCenter",
    "OSLog",
    "OpenAL",
    "OpenGLES",
    "PDFKit",
    "PHASE",
    "PassKit",
    "PencilKit",
    "Photos",
    "PhotosUI",
    "PushKit",
    "QuartzCore",
    "QuickLook",
    "QuickLookThumbnailing",
    "RealityFoundation",
    "RealityKit",
    "ReplayKit",
    "SafariServices",
    "SceneKit",
    "ScreenTime",
    "Security",
    "SensorKit",
    "ShazamKit",
    "Social",
    "SoundAnalysis",
    "Speech",
    "SpriteKit",
    "StoreKit",
    "SwiftUI",
    "SystemConfiguration",
    "TabularData",
    "Twitter",
    "UIKit",
    "UniformTypeIdentifiers",
    "UserNotifications",
    "UserNotificationsUI",
    "VideoSubscriberAccount",
    "VideoToolbox",
    "Vision",
    "VisionKit",
    "WatchConnectivity",
    "WebKit",
    "WidgetKit",
    "AppleTextureEncoder",
    "Compression",
    "Darwin",
    "asl",
    "dnssd",
    "os",
    "os_object",
    "os_workgroup",
    "libkern",
    "notify",
    "zlib",
    "SQLite3",
}

APPLE_TEST_FRAMEWORKS = {
    "XCTest",
}


# These modules require specific handling, as they do not have an umbrella
# header that matches the module name, as typical Apple frameworks do.
APPLE_SYSTEM_MODULE_OVERRIDES = {
    "Dispatch": ("dispatch", ("dispatch.h",)),
    "ObjectiveC": ("objc", ("runtime.h",)),
}


def write_imports_for_headers(out: TextIO, prefix: str, headers: Iterable[str]) -> None:
    for header in headers:
        print(f"#import <{prefix}/{header}>", file=out)


def write_imports_for_modules(
    out: TextIO,
    postprocessing_module_name: str,
    modules: Iterable[str],
    deps: Dict[str, Iterable[str]],
) -> None:
    # We only include the traditional textual imports when modules are disabled, so
    # that the behavior with modules enabled is identical to the behavior without
    # the postprocessing.
    print("#else", file=out)
    for module in modules:
        if headers := deps.get(module):
            write_imports_for_headers(out, module, headers)
        elif override := APPLE_SYSTEM_MODULE_OVERRIDES.get(module):
            write_imports_for_headers(out, override[0], override[1])
        elif module in APPLE_SYSTEM_MODULES or module in APPLE_TEST_FRAMEWORKS:
            # When we don't have an explicit override for the module, we use the module's
            # name as an umbrella header. This is used for typical Apple frameworks like
            # Foundation and UIKit.
            write_imports_for_headers(out, module, (f"{module}.h",))
        else:
            print(
                f"""
The module "{module}" was imported as a dependency of Swift code in "{postprocessing_module_name}", but could not be mapped to a list of header imports by Buck's Swift header postprocessing. There are two possibilities:

1. If "{module}" is an internal library, it is likely that the exported_deps of "{postprocessing_module_name}" are incorrect. Try fixing them manually or with "arc fixmydeps". This is the most likely issue.

2. If "{module}" is a system (Apple) framework, the list of Apple system modules in {os.path.basename(__file__)} is out-of-date. There is a command to fix it in that file. This issue is unlikely.
""",
                file=sys.stderr,
            )
            sys.exit(1)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("header")
    parser.add_argument("deps")
    parser.add_argument("out")
    args = parser.parse_args()

    with open(args.deps) as f:
        deps = json.load(f)

    # Strips the suffix from the header name, leaving us with just the name
    # of the module that we are postprocessing the header for. This is used
    # for error reporting.
    postprocessing_module_name = os.path.basename(args.header).split("-")[0]

    # The Swift compiler's output looks like this:
    #
    # #if __has_feature(modules)
    # #if __has_warning("-Watimport-in-framework-header")
    # #pragma clang diagnostic ignored "-Watimport-in-framework-header"
    # #endif
    # @import ModuleA;
    # @import ModuleB;
    # @import ModuleC;
    # #endif
    #
    # The implementation here balances being somewhat flexible to changes to the compiler's
    # output, unlikely though they may be, with avoiding adding too much complexity and getting
    # too close to implementing a full parser for Objective-C un-preprocessed header files.

    with open(args.header) as header, open(args.out, "w") as out:
        # When this is None, it means that we are still searching for the start of the conditional
        # @import block in the generated header.
        modules = None
        # The Swift compiler emits an additional #if gate inside the conditional @import block, so
        # we need to track whether we're in a further nested conditional so that we know when the
        # main conditional block has ended.
        if_level = 0

        for line in header:
            line = line.rstrip("\n")
            # When the modules has not been set, we are still searching for the start of the
            # modules @import section.
            if modules is None:
                if line == "#if __has_feature(modules)":
                    modules = []
                    if_level = 1
            else:
                if line.startswith("@import"):
                    # Splitting on:
                    #   "@import ": to separate from the @import.
                    #   Semicolon and period: to separate the main module name from submodules or EOL.
                    # The module name will then be the first item.
                    modules.append(re.split(r"@import |[;.]", line)[1])
                elif line.startswith("#if"):
                    # This allows us to handle the Clang diagnostic #if block that the compiler inserts
                    # within the main #if block for modules.
                    if_level += 1
                elif line.startswith("#endif"):
                    if_level -= 1
                    if if_level == 0:
                        write_imports_for_modules(
                            out,
                            postprocessing_module_name,
                            modules,
                            deps,
                        )
                        modules = None
            print(line, file=out)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Create a bootstrapper pex for inplace python binaries

This script:
    - Writes out a bootstrapper pex script that knows where this symlink tree is,
      and uses it, along with the provided entry point to run the python script.
      It does this by replacing a few special strings like <MODULES_DIR> and
      <MAIN_MODULE>

A full usage might be something like this:

$ cat template.in
(see prelude/python/run_inplace_lite.py.in)
$ ./make_pex_inplace.py  \\
    --template prelude/python/run_inplace.py.in \\
    # These two args create the hashbang for the bootstrapper script \\
    --python="/usr/bin/python3" \\
    --python-interpreter-flags="-Es" \\
    # This is based on the path in dests. This is the module that gets executed \\
    # to start program execution \\
    --entry-point=lib.foo  \\
    --output=bin.pex \\
    # This is the symlink tree \\
    --modules-dir=bin__link-tree
$ ./bin.pex
...
"""

import argparse
import errno
import os
import platform
import stat
from pathlib import Path


def parse_args() -> argparse.Namespace:
    # TODO(nmj): Go back and verify all of the various flags that make_xar
    #                 takes, and standardize on that so that the calling convention
    #                 is the same regardless of the "make_X" binary that's used.
    parser = argparse.ArgumentParser(
        description=(
            "Create a python inplace binary, writing a symlink tree to a directory, "
            "and a bootstrapper pex file to file"
        ),
        fromfile_prefix_chars="@",
    )
    parser.add_argument(
        "--template",
        required=True,
        type=Path,
        help="The template file for the .pex bootstrapper script",
    )
    parser.add_argument(
        "--template-lite",
        required=True,
        type=Path,
        help="The template file for the .pex bootstrapper script, if it's simple",
    )

    parser.add_argument(
        "--preload",
        type=Path,
        dest="preload_libraries",
        action="append",
        default=[],
        help="A list of native libraries to add to LD_PRELOAD",
    )
    parser.add_argument(
        "--python",
        required=True,
        help="The python binary to put in the bootstrapper hashbang",
    )
    parser.add_argument(
        "--host-python",
        required=True,
        help="The host python binary to use to e.g. compiling bytecode",
    )
    parser.add_argument(
        "--python-interpreter-flags",
        default="-Es",
        help="The interpreter flags for the hashbang",
    )
    parser.add_argument(
        "--entry-point", required=True, help="The main module to execute"
    )
    parser.add_argument(
        "--modules-dir",
        required=True,
        type=Path,
        help="The link tree directory to use at runtime",
    )
    parser.add_argument(
        "--use-lite",
        help="Whether to use the lite template",
        action="store_true",
    )
    parser.add_argument(
        "output",
        type=Path,
        help="Where to write the bootstrapper script to",
    )
    parser.add_argument(
        "--native-libs-env-var",
        default=(
            "DYLD_LIBRARY_PATH" if platform.system() == "Darwin" else "LD_LIBRARY_PATH"
        ),
        help="The dynamic loader env used to find native library deps",
    )
    # Compatibility with existing make_par scripts
    parser.add_argument("--passthrough", action="append", default=[])

    return parser.parse_args()


def write_bootstrapper(args: argparse.Namespace) -> None:
    """Write the .pex bootstrapper script using a template"""

    template = args.template_lite if args.use_lite else args.template
    with open(template, "r", encoding="utf8") as fin:
        data = fin.read()

    # Because this can be invoked from other directories, find the relative path
    # from this .par to the modules dir, and use that.
    relative_modules_dir = os.path.relpath(args.modules_dir, args.output.parent)

    # TODO(nmj): Remove this hack. So, if arg0 in your shebang is a bash script
    #                 (like /usr/local/fbcode/platform007/bin/python3.7 on macs is)
    #                 OSX just sort of ignores it and tries to run your thing with
    #                 the current shell. So, we hack in /usr/bin/env in the front
    #                 for now, and let it do the lifting. OSX: Bringing you the best
    #                 of 1980s BSD in 2021...
    #                 Also, make sure we add PYTHON_INTERPRETER_FLAGS back. We had to
    #                 exclude it for now, because linux doesn't like multiple args
    #                 after /usr/bin/env

    ld_preload = "None"
    if args.preload_libraries:
        ld_preload = repr(":".join(p.name for p in args.preload_libraries))

    new_data = data.replace("<PYTHON>", "/usr/bin/env " + str(args.python))
    new_data = new_data.replace("<PYTHON_INTERPRETER_FLAGS>", "")
    # new_data = new_data.replace(
    #    "<PYTHON_INTERPRETER_FLAGS>", args.python_interpreter_flags
    # )
    new_data = new_data.replace("<MODULES_DIR>", str(relative_modules_dir))
    new_data = new_data.replace("<MAIN_MODULE>", args.entry_point)

    # Things that are only required for the full template
    new_data = new_data.replace("<NATIVE_LIBS_ENV_VAR>", args.native_libs_env_var)
    new_data = new_data.replace("<NATIVE_LIBS_DIR>", repr(relative_modules_dir))
    new_data = new_data.replace("<NATIVE_LIBS_PRELOAD_ENV_VAR>", "LD_PRELOAD")
    new_data = new_data.replace("<NATIVE_LIBS_PRELOAD>", ld_preload)

    args.output.parent.mkdir(parents=True, exist_ok=True)
    with open(args.output, "w", encoding="utf8") as fout:
        fout.write(new_data)
    mode = os.stat(args.output).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    os.chmod(args.output, mode)


def main() -> None:
    args = parse_args()
    write_bootstrapper(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Compile the given Go sources into a Go package.

Example:

 $ ./compile_wrapper.py \
       --compiler compile \
       --assember assemble \
       --output srcs.txt src/dir/

"""

# pyre-unsafe

import argparse
import os
import subprocess
import sys
from pathlib import Path
from typing import List


def _compile(compile_prefix: List[str], output: Path, srcs: List[Path]):
    cmd = []
    cmd.extend(compile_prefix)
    cmd.append("-trimpath={}".format(os.getcwd()))
    cmd.append("-o")
    cmd.append(output)
    cmd.extend(srcs)
    subprocess.check_call(cmd)


def _pack(pack_prefix: List[str], output: Path, items: List[Path]):
    cmd = []
    cmd.extend(pack_prefix)
    cmd.append("r")
    cmd.append(output)
    cmd.extend(items)
    subprocess.check_call(cmd)


def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--compiler", action="append", default=[])
    parser.add_argument("--assembler", action="append", default=[])
    parser.add_argument("--packer", action="append", default=[])
    parser.add_argument("--output", required=True, type=Path)
    parser.add_argument("srcs", type=Path, nargs="*")
    args = parser.parse_args(argv[1:])

    # If there's no srcs, just leave an empty file.
    if not args.srcs:
        args.output.touch()
        return

    go_files = [s for s in args.srcs if s.suffix == ".go"]
    s_files = [s for s in args.srcs if s.suffix == ".s"]
    o_files = [s for s in args.srcs if s.suffix == ".o"]

    if go_files:
        compile_prefix = []
        compile_prefix.extend(args.compiler)

        # If we have assembly files, generate the symabi file to compile against.
        if s_files:
            symabis = args.output.with_suffix(".symabis")
            _compile(args.assembler + ["-gensymabis"], symabis, s_files)
            compile_prefix.extend(["-symabis", symabis])

        _compile(compile_prefix, args.output, go_files)

    else:
        args.output.touch()

    # If there are assembly files, assemble them to an object and add into the
    # output archive.
    if s_files:
        s_object = args.output.with_suffix(".o")
        _compile(args.assembler, s_object, s_files)
        o_files.append(s_object)

    if o_files:
        _pack(args.packer, args.output, o_files)


sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import json
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "env_file", help="A JSON file containing the environment to inject"
    )
    parser.add_argument("executable")
    parser.add_argument("args", nargs="*")

    args = parser.parse_args()

    with open(args.env_file) as env_file:
        env_from_file = json.load(env_file)
        env = {**os.environ, **env_from_file}

    os.execve(args.executable, [args.executable, *args.args], env)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

"""
Example usage:
$ cat inputs.manifest
[["foo.py", "input/foo.py", "//my_rule:foo"]]
$ compile.py --output=out-dir --ignore-errors inputs.manifest
$ find out-dir -type f
out-dir/foo.pyc
"""

# pyre-unsafe

import argparse
import errno
import json
import os
import py_compile
import sys


if sys.version_info[0] == 3:
    import importlib

    DEFAULT_FORMAT = importlib.util.cache_from_source("{pkg}/{name}.py")
else:
    DEFAULT_FORMAT = "{pkg}/{name}.pyc"


def get_py_path(module):
    return module.replace(".", os.sep) + ".py"


def get_pyc_path(module, fmt):
    try:
        package, name = module.rsplit(".", 1)
    except ValueError:
        package, name = "", module
    parts = fmt.split(os.sep)
    for idx in range(len(parts)):
        if parts[idx] == "{pkg}":
            parts[idx] = package.replace(".", os.sep)
        elif parts[idx].startswith("{name}"):
            parts[idx] = parts[idx].format(name=name)
    return os.path.join(*parts)


def _mkdirs(dirpath):
    try:
        os.makedirs(dirpath)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def main(argv):
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("-o", "--output", required=True)
    parser.add_argument("-f", "--format", default=DEFAULT_FORMAT)
    parser.add_argument("-i", "--ignore-errors", action="store_true")
    parser.add_argument("manifests", nargs="*")
    args = parser.parse_args(argv[1:])

    _mkdirs(args.output)

    for manifest_path in args.manifests:
        with open(manifest_path) as mf:
            manifest = json.load(mf)
        for dst, src, _ in manifest:
            # This is going to try to turn a path into a Python module, so
            # reduce the scope for bugs in get_pyc_path by normalizing first.
            dst = os.path.normpath(dst)
            # We only care about python sources.
            base, ext = os.path.splitext(dst)
            if ext != ".py":
                continue
            module = base.replace(os.sep, ".")
            pyc = os.path.join(args.output, get_pyc_path(module, args.format))
            _mkdirs(os.path.dirname(pyc))
            py_compile.compile(
                src,
                cfile=pyc,
                dfile=get_py_path(module),
                doraise=not args.ignore_errors,
                invalidation_mode=py_compile.PycInvalidationMode.UNCHECKED_HASH,
            )


sys.exit(main(sys.argv))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import sys
import zipfile


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--src", required=True, help="File to extract")
    parser.add_argument("--dst", required=True, help="Output directory")

    return parser.parse_args()


def do_unzip(src, dst):
    z = zipfile.ZipFile(src)
    z.extractall(dst)


def main():
    args = _parse_args()
    print("Source zip is: {}".format(args.src), file=sys.stderr)
    print("Output destination is: {}".format(args.dst), file=sys.stderr)
    do_unzip(args.src, args.dst)


if __name__ == "__main__":
    main()



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

# pyre-unsafe

import json
import shutil
import subprocess
import sys


def rewrite_dep_file(always_used_files_path, used_classes_path, dst_path):
    """
    Convert a used_classes.json to a depfile suitable for use by Buck2. The files we
    rewrite are JSON where the keys are the jars that were used.
    """
    shutil.copyfile(always_used_files_path, dst_path)

    with open(used_classes_path) as f:
        used_classes_body = f.read()

    used_classes_map = json.loads(used_classes_body)

    with open(dst_path, "a") as f:
        f.write("\n")
        f.write("\n".join(used_classes_map.keys()))


def main():
    """
    First argument is a file containing a list of files that should be put directly
    into the dep file.
    Second argument is a "used_classes.json" file.
    Third argument is where the dep file should be written.
    The command follows the third argument.
    """
    ret = subprocess.call(sys.argv[4:])
    if ret == 0:
        rewrite_dep_file(sys.argv[1], sys.argv[2], sys.argv[3])
    sys.exit(ret)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under both the MIT license found in the
# LICENSE-MIT file in the root directory of this source tree and the Apache
# License, Version 2.0 found in the LICENSE-APACHE file in the root directory
# of this source tree.

import argparse
import json
import os
import sys
from typing import List


def main(argv: List[str]) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("--origin", help="description of source origin")
    parser.add_argument("--prefix", help="prefix to prepend to destinations")
    parser.add_argument("extracted", help="path to directory of sources")
    args = parser.parse_args(argv[1:])

    entries = []
    for root, _, files in os.walk(args.extracted):
        for name in files:
            path = os.path.join(root, name)
            dest = os.path.relpath(path, args.extracted)
            if args.prefix is not None:
                dest = os.path.join(args.prefix, dest)
            entry = [dest, path]
            if args.origin is not None:
                entry.append(args.origin)
            entries.append(entry)

    json.dump(entries, args.output, indent=2, sort_keys=True)


sys.exit(main(sys.argv))
